"""
Functions for integrating population frequency spectra.
"""

import logging
logger = logging.getLogger('Integration')

# Note that these functions have all be written for xx=yy=zz, so the grids are
# identical in each direction. That's not essential here, though. The reason we
# do it is because later we'll want to extrapolate, and it's not obvious how to
# do so if the grids differ.
#
# Note that we could do different grids, if we solved a more complex
# extrapolation polynomial. For example:
#    val = a + b x + c x^2 + d y + e y^2 + f x y
# Where x and y denote the injection points for x and y axes.

# Also note that we have tested using multiprocessing to split up calls to
# the C integration methods. This appears to give no speedup (or even cause a
# slowdown). I suspect this is due to the cost of copying phi back and forth
# between subprocesses. I tried using shared memory to fix this, but was
# unsuccessful.

#: Controls use of Chang and Cooper's delj trick, which seems to lower accuracy.
use_delj_trick = False

import numpy
from numpy import newaxis as nuax

cuda_enabled = False
from . import Misc, Numerics, tridiag
from . import integration_c as int_c

#: Controls timestep for integrations. This is a reasonable default for
#: gridsizes of ~60. See set_timescale_factor for better control.
timescale_factor = 1e-3

#: Whether to use old timestep method, which is old_timescale_factor * dx[0].
use_old_timestep = False
#: Factor for told timestep method.
old_timescale_factor = 0.1

def set_timescale_factor(pts, factor=10):
    """
    Controls the fineness of timesteps during integration.

    The timestep will be proportional to Numerics.default_grid(pts)[1]/factor.
    Typically, pts should be set to the *largest* number of grid points used in
    extrapolation. 
    
    An adjustment factor of 10 typically results in acceptable accuracy. It may
    be desirable to increase this factor, particularly when population sizes
    are changing continously and rapidly.
    """
    # Implementation note: This cannot be easily be set automatically, because
    # the integration doesn't know whether its results will be used in an
    # extrapolation.
    global timescale_factor
    logger.warn('set_timescale_factor has been deprecated, as it may be too '
                'conservative (and thus slow) in choosing timesteps. If you '
                'wish to take smaller timesteps for accuracy (particularly for '
                'a very quickly growing population), manually set '
                'dadi.Integration.timescale_factor to a smaller value. '
                '(Current value is %g.)' % timescale_factor)
    timescale_factor = Numerics.default_grid(pts)[1]/factor

def _inject_mutations_1D(phi, dt, xx, theta0):
    """
    Inject novel mutations for a timestep.
    """
    phi[1] += dt/xx[1] * theta0/2 * 2/(xx[2] - xx[0])
    return phi
def _inject_mutations_2D(phi, dt, xx, yy, theta0, frozen1, frozen2,
                         nomut1, nomut2):
    """
    Inject novel mutations for a timestep.
    """
    # Population 1
    if not frozen1 and not nomut1:
        phi[1,0] += dt/xx[1] * theta0/2 * 4/((xx[2] - xx[0]) * yy[1])
    # Population 2
    if not frozen2 and not nomut2:
        phi[0,1] += dt/yy[1] * theta0/2 * 4/((yy[2] - yy[0]) * xx[1])
    return phi
def _inject_mutations_3D(phi, dt, xx, yy, zz, theta0, frozen1, frozen2, frozen3):
    """
    Inject novel mutations for a timestep.
    """
    # Population 1
    # Normalization based on the multi-dimensional trapezoid rule is 
    # implemented                      ************** here ***************
    if not frozen1:
        phi[1,0,0] += dt/xx[1] * theta0/2 * 8/((xx[2] - xx[0]) * yy[1] * zz[1])
    # Population 2
    if not frozen2:
        phi[0,1,0] += dt/yy[1] * theta0/2 * 8/((yy[2] - yy[0]) * xx[1] * zz[1])
    # Population 3
    if not frozen3:
        phi[0,0,1] += dt/zz[1] * theta0/2 * 8/((zz[2] - zz[0]) * xx[1] * yy[1])
    return phi
def _inject_mutations_4D(phi, dt, xx, yy, zz, aa, theta0, frozen1, frozen2, frozen3, frozen4):
    """
    Inject novel mutations for a timestep.
    """
    # Population 1
    # Normalization based on the multi-dimensional trapezoid rule is 
    # implemented                      ************** here ***************
    if not frozen1:
        phi[1,0,0,0] += dt/xx[1] * theta0/2 * 16/((xx[2] - xx[0]) * yy[1] * zz[1] * aa[1])
    # Population 2
    if not frozen2:
        phi[0,1,0,0] += dt/yy[1] * theta0/2 * 16/((yy[2] - yy[0]) * xx[1] * zz[1] * aa[1])
    # Population 3
    if not frozen3:
        phi[0,0,1,0] += dt/zz[1] * theta0/2 * 16/((zz[2] - zz[0]) * xx[1] * yy[1] * aa[1])
    # Population 4
    if not frozen4:
        phi[0,0,0,1] += dt/aa[1] * theta0/2 * 16/((aa[2] - aa[0]) * xx[1] * yy[1] * zz[1])
    return phi

def _inject_mutations_5D(phi, dt, xx, yy, zz, aa, bb, theta0, frozen1, frozen2, frozen3, frozen4, frozen5):
    """
    Inject novel mutations for a timestep.
    """
    # Population 1
    # Normalization based on the multi-dimensional trapezoid rule is 
    # implemented                      ************** here ***************
    if not frozen1:
        phi[1,0,0,0,0] += dt/xx[1] * theta0/2 * 32/((xx[2] - xx[0]) * yy[1] * zz[1] * aa[1] * bb[1])
    # Population 2
    if not frozen2:
        phi[0,1,0,0,0] += dt/yy[1] * theta0/2 * 32/((yy[2] - yy[0]) * xx[1] * zz[1] * aa[1] * bb[1])
    # Population 3
    if not frozen3:
        phi[0,0,1,0,0] += dt/zz[1] * theta0/2 * 32/((zz[2] - zz[0]) * xx[1] * yy[1] * aa[1] * bb[1])
    # Population 4
    if not frozen4:
        phi[0,0,0,1,0] += dt/aa[1] * theta0/2 * 32/((aa[2] - aa[0]) * xx[1] * yy[1] * zz[1] * bb[1])
    # Population 5
    if not frozen5:
        phi[0,0,0,0,1] += dt/bb[1] * theta0/2 * 32/((bb[2] - bb[0]) * xx[1] * yy[1] * zz[1] * aa[1])
    return phi

def _compute_dt(dx, nu, ms, gamma, h):
    """
    Compute the appropriate timestep given the current demographic params.

    This is based on the maximum V or M expected in this direction. The
    timestep is scaled such that if the params are rescaled correctly by a
    constant, the exact same integration happens. (This is equivalent to
    multiplying the eqn through by some other 2N...)
    """
    if use_old_timestep:
        return old_timescale_factor * dx[0]

    # These are the maxima for V_func and M_func over the domain
    # For h != 0.5, the maximum of M_func is not easy analytically. It is close
    # to the 0.5 or 0.25 value, though, so we use those as an approximation.

    # It might seem natural to scale dt based on dx[0]. However, testing has
    # shown that extrapolation is much more reliable when the same timesteps
    # are used in evaluations at different grid sizes.
    maxVM = max(0.25/nu, sum(ms),\
                abs(gamma) * 2*max(numpy.abs(h + (1-2*h)*0.5) * 0.5*(1-0.5),
                                   numpy.abs(h + (1-2*h)*0.25) * 0.25*(1-0.25)))
    if maxVM > 0:
        dt = timescale_factor / maxVM
    else:
        dt = numpy.inf
    if dt == 0:
        raise ValueError('Timestep is zero. Values passed in are nu=%f, ms=%s,'
                         'gamma=%f, h=%f.' % (nu, str(ms), gamma, h))
    return dt

def one_pop(phi, xx, T, nu=1, gamma=0, h=0.5, theta0=1.0, initial_t=0, 
            frozen=False, beta=1):
    """
    Integrate a 1-dimensional phi foward.

    phi: Initial 1-dimensional phi
    xx: Grid upon (0,1) overwhich phi is defined.

    nu, gamma, and theta0 may be functions of time.
    nu: Population size
    gamma: Selection coefficient on *all* segregating alleles
    h: Dominance coefficient. h = 0.5 corresponds to genic selection. 
       Heterozygotes have fitness 1+2sh and homozygotes have fitness 1+2s.
    theta0: Propotional to ancestral size. Typically constant.
    beta: Breeding ratio, beta=Nf/Nm.

    T: Time at which to halt integration
    initial_t: Time at which to start integration. (Note that this only matters
               if one of the demographic parameters is a function of time.)

    frozen: If True, population is 'frozen' so that it does not change.
            In the one_pop case, this is equivalent to not running the
            integration at all.
    """
    phi = phi.copy()

    # For a one population integration, freezing means just not integrating.
    if frozen:
        return phi

    if T - initial_t == 0:
        return phi
    elif T - initial_t < 0:
        raise ValueError('Final integration time T (%f) is less than '
                         'intial_time (%f). Integration cannot be run '
                         'backwards.' % (T, initial_t))

    vars_to_check = (nu, gamma, h, theta0, beta)
    if numpy.all([numpy.isscalar(var) for var in vars_to_check]):
        return _one_pop_const_params(phi, xx, T, nu, gamma, h, theta0, 
                                     initial_t, beta)

    nu_f = Misc.ensure_1arg_func(nu)
    gamma_f = Misc.ensure_1arg_func(gamma)
    h_f = Misc.ensure_1arg_func(h)
    theta0_f = Misc.ensure_1arg_func(theta0)
    beta_f = Misc.ensure_1arg_func(beta)

    current_t = initial_t
    nu, gamma, h = nu_f(current_t), gamma_f(current_t), h_f(current_t)
    beta = beta_f(current_t)
    dx = numpy.diff(xx)
    while current_t < T:
        dt = _compute_dt(dx,nu,[0],gamma,h)
        this_dt = min(dt, T - current_t)

        # Because this is an implicit method, I need the *next* time's params.
        # So there's a little inconsistency here, in that I'm estimating dt
        # using the last timepoints nu,gamma,h.
        next_t = current_t + this_dt
        nu, gamma, h = nu_f(next_t), gamma_f(next_t), h_f(next_t)
        beta = beta_f(next_t)
        theta0 = theta0_f(next_t)

        if numpy.any(numpy.less([T,nu,theta0], 0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if numpy.any(numpy.equal([nu], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        _inject_mutations_1D(phi, this_dt, xx, theta0)
        # Do each step in C, since it will be faster to compute the a,b,c
        # matrices there.
        phi = int_c.implicit_1Dx(phi, xx, nu, gamma, h, beta, this_dt, 
                                 use_delj_trick=use_delj_trick)
        current_t = next_t
    return phi

def two_pops(phi, xx, T, nu1=1, nu2=1, m12=0, m21=0, gamma1=0, gamma2=0,
             h1=0.5, h2=0.5, theta0=1, initial_t=0, frozen1=False,
             frozen2=False, nomut1=False, nomut2=False, enable_cuda_cached=False):
    """
    Integrate a 2-dimensional phi foward.

    phi: Initial 2-dimensional phi
    xx: 1-dimensional grid upon (0,1) overwhich phi is defined. It is assumed
        that this grid is used in all dimensions.

    nu's, gamma's, m's, and theta0 may be functions of time.
    nu1,nu2: Population sizes
    gamma1,gamma2: Selection coefficients on *all* segregating alleles
    h1,h2: Dominance coefficients. h = 0.5 corresponds to genic selection.
    m12,m21: Migration rates. Note that m12 is the rate *into 1 from 2*.
    theta0: Propotional to ancestral size. Typically constant.

    T: Time at which to halt integration
    initial_t: Time at which to start integration. (Note that this only matters
               if one of the demographic parameters is a function of time.)

    frozen1,frozen2: If True, the corresponding population is "frozen" in time
                     (no new mutations and no drift), so the resulting spectrum
                     will correspond to an ancient DNA sample from that
                     population.

    nomut1,nomut2: If True, no new mutations will be introduced into the
                   given population.

    enable_cuda_cached: If True, enable CUDA integration with slower constant
                       parameter method. Likely useful only for benchmarking.

    Note: Generalizing to different grids in different phi directions is
          straightforward. The tricky part will be later doing the extrapolation
          correctly.
    """
    phi = phi.copy()

    if T - initial_t == 0:
        return phi
    elif T - initial_t < 0:
        raise ValueError('Final integration time T (%f) is less than '
                         'intial_time (%f). Integration cannot be run '
                         'backwards.' % (T, initial_t))

    if (frozen1 or frozen2) and (m12 != 0 or m21 != 0):
        raise ValueError('Population cannot be frozen and have non-zero '
                         'migration to or from it.')

    vars_to_check = [nu1,nu2,m12,m21,gamma1,gamma2,h1,h2,theta0]
    if False and numpy.all([numpy.isscalar(var) for var in vars_to_check]):
        # Constant integration with CUDA turns out to be slower,
        # so we only use it in specific circumsances.
        if not cuda_enabled or (cuda_enabled and enable_cuda_cached):
            return _two_pops_const_params(phi, xx, T, nu1, nu2, m12, m21,
                    gamma1, gamma2, h1, h2, theta0, initial_t,
                    frozen1, frozen2, nomut1, nomut2)
    yy = xx

    nu1_f = Misc.ensure_1arg_func(nu1)
    nu2_f = Misc.ensure_1arg_func(nu2)
    m12_f = Misc.ensure_1arg_func(m12)
    m21_f = Misc.ensure_1arg_func(m21)
    gamma1_f = Misc.ensure_1arg_func(gamma1)
    gamma2_f = Misc.ensure_1arg_func(gamma2)
    h1_f = Misc.ensure_1arg_func(h1)
    h2_f = Misc.ensure_1arg_func(h2)
    theta0_f = Misc.ensure_1arg_func(theta0)

    if cuda_enabled:
        import dadi.cuda
        phi = dadi.cuda.Integration._two_pops_temporal_params(phi, xx, T, initial_t,
                nu1_f, nu2_f, m12_f, m21_f, gamma1_f, gamma2_f, h1_f, h2_f, theta0_f, 
                frozen1, frozen2, nomut1, nomut2)
        return phi

    current_t = initial_t
    nu1,nu2 = nu1_f(current_t), nu2_f(current_t)
    m12,m21 = m12_f(current_t), m21_f(current_t)
    gamma1,gamma2 = gamma1_f(current_t), gamma2_f(current_t)
    h1,h2 = h1_f(current_t), h2_f(current_t)
    dx,dy = numpy.diff(xx),numpy.diff(yy)
    while current_t < T:
        dt = min(_compute_dt(dx,nu1,[m12],gamma1,h1),
                 _compute_dt(dy,nu2,[m21],gamma2,h2))
        this_dt = min(dt, T - current_t)

        next_t = current_t + this_dt

        nu1,nu2 = nu1_f(next_t), nu2_f(next_t)
        m12,m21 = m12_f(next_t), m21_f(next_t)
        gamma1,gamma2 = gamma1_f(next_t), gamma2_f(next_t)
        h1,h2 = h1_f(next_t), h2_f(next_t)
        theta0 = theta0_f(next_t)

        if numpy.any(numpy.less([T,nu1,nu2,m12,m21,theta0], 0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if numpy.any(numpy.equal([nu1,nu2], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        _inject_mutations_2D(phi, this_dt, xx, yy, theta0, frozen1, frozen2,
                             nomut1, nomut2)
        if not frozen1:
            phi = int_c.implicit_2Dx(phi, xx, yy, nu1, m12, gamma1, h1,
                                     this_dt, use_delj_trick)
        if not frozen2:
            phi = int_c.implicit_2Dy(phi, xx, yy, nu2, m21, gamma2, h2,
                                     this_dt, use_delj_trick)

        current_t = next_t
    return phi

def three_pops(phi, xx, T, nu1=1, nu2=1, nu3=1,
               m12=0, m13=0, m21=0, m23=0, m31=0, m32=0,
               gamma1=0, gamma2=0, gamma3=0, h1=0.5, h2=0.5, h3=0.5,
               theta0=1, initial_t=0, frozen1=False, frozen2=False,
               frozen3=False, enable_cuda_cached=False):
    """
    Integrate a 3-dimensional phi foward.

    phi: Initial 3-dimensional phi
    xx: 1-dimensional grid upon (0,1) overwhich phi is defined. It is assumed
        that this grid is used in all dimensions.

    nu's, gamma's, m's, and theta0 may be functions of time.
    nu1,nu2,nu3: Population sizes
    gamma1,gamma2,gamma3: Selection coefficients on *all* segregating alleles
    h1,h2,h3: Dominance coefficients. h = 0.5 corresponds to genic selection.
    m12,m13,m21,m23,m31,m32: Migration rates. Note that m12 is the rate 
                             *into 1 from 2*.
    theta0: Propotional to ancestral size. Typically constant.

    T: Time at which to halt integration
    initial_t: Time at which to start integration. (Note that this only matters
               if one of the demographic parameters is a function of time.)

    enable_cuda_cached: If True, enable CUDA integration with slower constant
                       parameter method. Likely useful only for benchmarking.

    Note: Generalizing to different grids in different phi directions is
          straightforward. The tricky part will be later doing the extrapolation
          correctly.
    """
    phi = phi.copy()

    if T - initial_t == 0:
        return phi
    elif T - initial_t < 0:
        raise ValueError('Final integration time T (%f) is less than '
                         'intial_time (%f). Integration cannot be run '
                         'backwards.' % (T, initial_t))


    if (frozen1 and (m12 != 0 or m21 != 0 or m13 !=0 or m31 != 0))\
       or (frozen2 and (m12 != 0 or m21 != 0 or m23 !=0 or m32 != 0))\
       or (frozen3 and (m13 != 0 or m31 != 0 or m23 !=0 or m32 != 0)):
        raise ValueError('Population cannot be frozen and have non-zero '
                         'migration to or from it.')

    vars_to_check = [nu1,nu2,nu3,m12,m13,m21,m23,m31,m32,gamma1,gamma2,
                     gamma3,h1,h2,h3,theta0]
    if False and numpy.all([numpy.isscalar(var) for var in vars_to_check]):
        if not cuda_enabled or (cuda_enabled and enable_cuda_cached):
            return _three_pops_const_params(phi, xx, T, nu1, nu2, nu3,
                                            m12, m13, m21, m23, m31, m32,
                                            gamma1, gamma2, gamma3, h1, h2, h3,
                                            theta0, initial_t,
                                            frozen1, frozen2, frozen3)
    zz = yy = xx

    nu1_f = Misc.ensure_1arg_func(nu1)
    nu2_f = Misc.ensure_1arg_func(nu2)
    nu3_f = Misc.ensure_1arg_func(nu3)
    m12_f = Misc.ensure_1arg_func(m12)
    m13_f = Misc.ensure_1arg_func(m13)
    m21_f = Misc.ensure_1arg_func(m21)
    m23_f = Misc.ensure_1arg_func(m23)
    m31_f = Misc.ensure_1arg_func(m31)
    m32_f = Misc.ensure_1arg_func(m32)
    gamma1_f = Misc.ensure_1arg_func(gamma1)
    gamma2_f = Misc.ensure_1arg_func(gamma2)
    gamma3_f = Misc.ensure_1arg_func(gamma3)
    h1_f = Misc.ensure_1arg_func(h1)
    h2_f = Misc.ensure_1arg_func(h2)
    h3_f = Misc.ensure_1arg_func(h3)
    theta0_f = Misc.ensure_1arg_func(theta0)

    if cuda_enabled:
        import dadi.cuda
        phi = dadi.cuda.Integration._three_pops_temporal_params(phi, xx, T, initial_t,
                nu1_f, nu2_f, nu3_f, m12_f, m13_f, m21_f, m23_f, m31_f, m32_f, 
                gamma1_f, gamma2_f, gamma3_f, h1_f, h2_f, h3_f, 
                theta0_f, frozen1, frozen2, frozen3)
        return phi

    current_t = initial_t
    nu1,nu2,nu3 = nu1_f(current_t), nu2_f(current_t), nu3_f(current_t)
    m12,m13 = m12_f(current_t), m13_f(current_t)
    m21,m23 = m21_f(current_t), m23_f(current_t)
    m31,m32 = m31_f(current_t), m32_f(current_t)
    gamma1,gamma2 = gamma1_f(current_t), gamma2_f(current_t)
    gamma3 = gamma3_f(current_t)
    h1,h2,h3 = h1_f(current_t), h2_f(current_t), h3_f(current_t)
    dx,dy,dz = numpy.diff(xx),numpy.diff(yy),numpy.diff(zz)
    while current_t < T:
        dt = min(_compute_dt(dx,nu1,[m12,m13],gamma1,h1),
                 _compute_dt(dy,nu2,[m21,m23],gamma2,h2),
                 _compute_dt(dz,nu3,[m31,m32],gamma3,h3))
        this_dt = min(dt, T - current_t)

        next_t = current_t + this_dt

        nu1,nu2,nu3 = nu1_f(next_t), nu2_f(next_t), nu3_f(next_t)
        m12,m13 = m12_f(next_t), m13_f(next_t)
        m21,m23 = m21_f(next_t), m23_f(next_t)
        m31,m32 = m31_f(next_t), m32_f(next_t)
        gamma1,gamma2 = gamma1_f(next_t), gamma2_f(next_t)
        gamma3 = gamma3_f(next_t)
        h1,h2,h3 = h1_f(next_t), h2_f(next_t), h3_f(next_t)
        theta0 = theta0_f(next_t)

        if numpy.any(numpy.less([T,nu1,nu2,nu3,m12,m13,m21,m23,m31,m32,theta0],
                                0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if numpy.any(numpy.equal([nu1,nu2,nu3], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        _inject_mutations_3D(phi, this_dt, xx, yy, zz, theta0,
                             frozen1, frozen2, frozen3)
        if not frozen1:
            phi = int_c.implicit_3Dx(phi, xx, yy, zz, nu1, m12, m13, 
                                     gamma1, h1, this_dt, use_delj_trick)
        if not frozen2:
            phi = int_c.implicit_3Dy(phi, xx, yy, zz, nu2, m21, m23, 
                                     gamma2, h2, this_dt, use_delj_trick)
        if not frozen3:
            phi = int_c.implicit_3Dz(phi, xx, yy, zz, nu3, m31, m32, 
                                     gamma3, h3, this_dt, use_delj_trick)

        current_t = next_t
    return phi

def four_pops(phi, xx, T, nu1=1, nu2=1, nu3=1, nu4=1,
              m12=0, m13=0, m14=0, m21=0, m23=0, m24=0, 
              m31=0, m32=0, m34=0, m41=0, m42=0, m43=0,
              gamma1=0, gamma2=0, gamma3=0, gamma4=0, 
              h1=0.5, h2=0.5, h3=0.5, h4=0.5,
              theta0=1, initial_t=0, frozen1=False, frozen2=False,
              frozen3=False, frozen4=False):
    """
    Integrate a 4-dimensional phi foward.

    phi: Initial 4-dimensional phi
    xx: 1-dimensional grid upon (0,1) overwhich phi is defined. It is assumed
        that this grid is used in all dimensions.

    nu's, gamma's, m's, and theta0 may be functions of time.
    nu1,nu2,nu3,nu4: Population sizes
    gamma1,gamma2,gamma3,gamma4: Selection coefficients on *all* segregating alleles
    h1,h2,h3,h4: Dominance coefficients. h = 0.5 corresponds to genic selection.
    m12,m13,m21,m23,m31,m32, ...: Migration rates. Note that m12 is the rate 
                             *into 1 from 2*.
    theta0: Proportional to ancestral size. Typically constant.

    T: Time at which to halt integration
    initial_t: Time at which to start integration. (Note that this only matters
               if one of the demographic parameters is a function of time.)

    enable_cuda_const: If True, enable CUDA integration with slower constant
                       parameter method. Likely useful only for benchmarking.

    Note: Generalizing to different grids in different phi directions is
          straightforward. The tricky part will be later doing the extrapolation
          correctly.
    """
    if T - initial_t == 0:
        return phi
    elif T - initial_t < 0:
        raise ValueError('Final integration time T (%f) is less than '
                         'intial_time (%f). Integration cannot be run '
                         'backwards.' % (T, initial_t))


    if (frozen1 and (m12 != 0 or m21 != 0 or m13 !=0 or m31 != 0 or m41 != 0 or m14 != 0))\
       or (frozen2 and (m12 != 0 or m21 != 0 or m23 != 0 or m32 != 0 or m24 != 0 or m42 != 0))\
       or (frozen3 and (m13 != 0 or m31 != 0 or m23 !=0 or m32 != 0 or m34 != 0 or m43 != 0))\
       or (frozen4 and (m14 != 0 or m41 != 0 or m24 !=0 or m42 != 0 or m34 != 0 or m43 != 0)):
        raise ValueError('Population cannot be frozen and have non-zero '
                         'migration to or from it.')
    aa = zz = yy = xx

    nu1_f, nu2_f = Misc.ensure_1arg_func(nu1), Misc.ensure_1arg_func(nu2)
    nu3_f, nu4_f = Misc.ensure_1arg_func(nu3), Misc.ensure_1arg_func(nu4)
    gamma1_f, gamma2_f = Misc.ensure_1arg_func(gamma1), Misc.ensure_1arg_func(gamma2)
    gamma3_f, gamma4_f = Misc.ensure_1arg_func(gamma3), Misc.ensure_1arg_func(gamma4)
    h1_f, h2_f = Misc.ensure_1arg_func(h1), Misc.ensure_1arg_func(h2)
    h3_f, h4_f = Misc.ensure_1arg_func(h3), Misc.ensure_1arg_func(h4)
    m12_f, m13_f, m14_f = Misc.ensure_1arg_func(m12), Misc.ensure_1arg_func(m13), Misc.ensure_1arg_func(m14)
    m21_f, m23_f, m24_f = Misc.ensure_1arg_func(m21), Misc.ensure_1arg_func(m23), Misc.ensure_1arg_func(m24)
    m31_f, m32_f, m34_f = Misc.ensure_1arg_func(m31), Misc.ensure_1arg_func(m32), Misc.ensure_1arg_func(m34)
    m41_f, m42_f, m43_f = Misc.ensure_1arg_func(m41), Misc.ensure_1arg_func(m42), Misc.ensure_1arg_func(m43)
    theta0_f = Misc.ensure_1arg_func(theta0)

    if cuda_enabled:
        import dadi.cuda
        phi = dadi.cuda.Integration._four_pops_temporal_params(phi, xx, T, initial_t,
                nu1_f, nu2_f, nu3_f, nu4_f, m12_f, m13_f, m14_f, m21_f, m23_f, m24_f, m31_f, m32_f, m34_f,
                m41_f, m42_f, m43_f, gamma1_f, gamma2_f, gamma3_f, gamma4_f, h1_f, h2_f, h3_f, h4_f,
                theta0_f, frozen1, frozen2, frozen3, frozen4)
        return phi

    current_t = initial_t
    nu1, nu2, nu3, nu4 = nu1_f(current_t), nu2_f(current_t), nu3_f(current_t), nu4_f(current_t)
    gamma1, gamma2, gamma3, gamma4 = gamma1_f(current_t), gamma2_f(current_t), gamma3_f(current_t), gamma4_f(current_t)
    h1, h2, h3, h4 = h1_f(current_t), h2_f(current_t), h3_f(current_t), h4_f(current_t)
    m12, m13, m14 = m12_f(current_t), m13_f(current_t), m14_f(current_t)
    m21, m23, m24 = m21_f(current_t), m23_f(current_t), m24_f(current_t)
    m31, m32, m34 = m31_f(current_t), m32_f(current_t), m34_f(current_t)
    m41, m42, m43 = m41_f(current_t), m42_f(current_t), m43_f(current_t)

    dx,dy,dz,da = numpy.diff(xx),numpy.diff(yy),numpy.diff(zz),numpy.diff(aa)
    while current_t < T:
        dt = min(_compute_dt(dx,nu1,[m12,m13,m14],gamma1,h1),
                 _compute_dt(dy,nu2,[m21,m23,m24],gamma2,h2),
                 _compute_dt(dz,nu3,[m31,m32,m34],gamma3,h3),
                 _compute_dt(da,nu4,[m41,m42,m43],gamma4,h4))
        this_dt = min(dt, T - current_t)

        next_t = current_t + this_dt

        nu1, nu2, nu3, nu4 = nu1_f(next_t), nu2_f(next_t), nu3_f(next_t), nu4_f(next_t)
        gamma1, gamma2, gamma3, gamma4 = gamma1_f(next_t), gamma2_f(next_t), gamma3_f(next_t), gamma4_f(next_t)
        h1, h2, h3, h4 = h1_f(next_t), h2_f(next_t), h3_f(next_t), h4_f(next_t)
        m12, m13, m14 = m12_f(next_t), m13_f(next_t), m14_f(next_t)
        m21, m23, m24 = m21_f(next_t), m23_f(next_t), m24_f(next_t)
        m31, m32, m34 = m31_f(next_t), m32_f(next_t), m34_f(next_t)
        m41, m42, m43 = m41_f(next_t), m42_f(next_t), m43_f(next_t)
        theta0 = theta0_f(next_t)

        if numpy.any(numpy.less([T,nu1,nu2,nu3,nu4,m12,m13,m14,m21,
                                 m23, m24, m31, m32, m34, m41, m42, m43, theta0],
                                0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if numpy.any(numpy.equal([nu1,nu2,nu3,nu4], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        _inject_mutations_4D(phi, this_dt, xx, yy, zz, aa, theta0,
                             frozen1, frozen2, frozen3, frozen4)
        if not frozen1:
            phi = int_c.implicit_4Dx(phi, xx, yy, zz, aa, nu1, m12, m13, m14,
                                     gamma1, h1, this_dt, use_delj_trick)
        if not frozen2:
            phi = int_c.implicit_4Dy(phi, xx, yy, zz, aa, nu2, m21, m23, m24,
                                     gamma2, h2, this_dt, use_delj_trick)
        if not frozen3:
            phi = int_c.implicit_4Dz(phi, xx, yy, zz, aa, nu3, m31, m32, m34,
                                     gamma3, h3, this_dt, use_delj_trick)
        if not frozen4:
            phi = int_c.implicit_4Da(phi, xx, yy, zz, aa, nu4, m41, m42, m43,
                                     gamma4, h4, this_dt, use_delj_trick)

        current_t = next_t
    return phi

def five_pops(phi, xx, T, nu1=1, nu2=1, nu3=1, nu4=1, nu5=1,
              m12=0, m13=0, m14=0, m15=0, m21=0, m23=0, m24=0, m25=0,
              m31=0, m32=0, m34=0, m35=0, m41=0, m42=0, m43=0, m45=0,
              m51=0, m52=0, m53=0, m54=0, 
              gamma1=0, gamma2=0, gamma3=0, gamma4=0, gamma5=0,
              h1=0.5, h2=0.5, h3=0.5, h4=0.5, h5=0.5,
              theta0=1, initial_t=0, frozen1=False, frozen2=False,
              frozen3=False, frozen4=False, frozen5=False):
    """
    Integrate a 5-dimensional phi foward.

    phi: Initial 5-dimensional phi
    xx: 1-dimensional grid upon (0,1) overwhich phi is defined. It is assumed
        that this grid is used in all dimensions.

    nu's, gamma's, m's, and theta0 may be functions of time.
    nu1,nu2,nu3,nu4,nu5: Population sizes
    gamma1,gamma2,gamma3,gamma4,gamma5: Selection coefficients on *all* segregating alleles
    h1,h2,h3,h4,h5: Dominance coefficients. h = 0.5 corresponds to genic selection.
    m12,m13,m21,m23,m31,m32, ...: Migration rates. Note that m12 is the rate 
                             *into 1 from 2*.
    theta0: Proportional to ancestral size. Typically constant.

    T: Time at which to halt integration
    initial_t: Time at which to start integration. (Note that this only matters
               if one of the demographic parameters is a function of time.)

    Note: Generalizing to different grids in different phi directions is
          straightforward. The tricky part will be later doing the extrapolation
          correctly.
    """
    if T - initial_t == 0:
        return phi
    elif T - initial_t < 0:
        raise ValueError('Final integration time T (%f) is less than '
                         'intial_time (%f). Integration cannot be run '
                         'backwards.' % (T, initial_t))

    if (frozen1 and (m12 != 0 or m21 != 0 or m13 !=0 or m31 != 0 or m41 != 0 or m14 != 0 or m15 != 0 or m51 != 0))\
       or (frozen2 and (m12 != 0 or m21 != 0 or m23 != 0 or m32 != 0 or m24 != 0 or m42 != 0 or m25 != 0 or m52 != 0))\
       or (frozen3 and (m13 != 0 or m31 != 0 or m23 != 0 or m32 != 0 or m34 != 0 or m43 != 0 or m35 != 0 or m53 !=0))\
       or (frozen4 and (m14 != 0 or m41 != 0 or m24 != 0 or m42 != 0 or m34 != 0 or m43 != 0 or m45 != 0 or m54 != 0))\
       or (frozen5 and (m15 != 0 or m51 != 0 or m25 != 0 or m52 != 0 or m35 != 0 or m53 != 0 or m45 != 0 or m54 != 0)):
        raise ValueError('Population cannot be frozen and have non-zero '
                         'migration to or from it.')
    bb = aa = zz = yy = xx

    nu1_f, nu2_f = Misc.ensure_1arg_func(nu1), Misc.ensure_1arg_func(nu2)
    nu3_f, nu4_f = Misc.ensure_1arg_func(nu3), Misc.ensure_1arg_func(nu4)
    nu5_f = Misc.ensure_1arg_func(nu5)
    gamma1_f, gamma2_f = Misc.ensure_1arg_func(gamma1), Misc.ensure_1arg_func(gamma2)
    gamma3_f, gamma4_f = Misc.ensure_1arg_func(gamma3), Misc.ensure_1arg_func(gamma4)
    gamma5_f = Misc.ensure_1arg_func(gamma5)
    h1_f, h2_f = Misc.ensure_1arg_func(h1), Misc.ensure_1arg_func(h2)
    h3_f, h4_f = Misc.ensure_1arg_func(h3), Misc.ensure_1arg_func(h4)
    h5_f = Misc.ensure_1arg_func(h5)
    m12_f, m13_f, m14_f, m15_f = Misc.ensure_1arg_func(m12), Misc.ensure_1arg_func(m13), Misc.ensure_1arg_func(m14), Misc.ensure_1arg_func(m15)
    m21_f, m23_f, m24_f, m25_f = Misc.ensure_1arg_func(m21), Misc.ensure_1arg_func(m23), Misc.ensure_1arg_func(m24), Misc.ensure_1arg_func(m25)
    m31_f, m32_f, m34_f, m35_f = Misc.ensure_1arg_func(m31), Misc.ensure_1arg_func(m32), Misc.ensure_1arg_func(m34), Misc.ensure_1arg_func(m35)
    m41_f, m42_f, m43_f, m45_f = Misc.ensure_1arg_func(m41), Misc.ensure_1arg_func(m42), Misc.ensure_1arg_func(m43), Misc.ensure_1arg_func(m45)
    m51_f, m52_f, m53_f, m54_f = Misc.ensure_1arg_func(m51), Misc.ensure_1arg_func(m52), Misc.ensure_1arg_func(m53), Misc.ensure_1arg_func(m54)
    theta0_f = Misc.ensure_1arg_func(theta0)

    if cuda_enabled:
        import dadi.cuda
        phi = dadi.cuda.Integration._five_pops_temporal_params(phi, xx, T, initial_t, 
            nu1_f, nu2_f, nu3_f, nu4_f, nu5_f,
            m12_f, m13_f, m14_f, m15_f, m21_f, m23_f, m24_f, m25_f, m31_f, m32_f, m34_f, m35_f,
            m41_f, m42_f, m43_f, m45_f, m51_f, m52_f, m53_f, m54_f, 
            gamma1_f, gamma2_f, gamma3_f, gamma4_f, gamma5_f,
            h1_f, h2_f, h3_f, h4_f, h5_f, theta0_f, frozen1, frozen2, frozen3, frozen4, frozen5)
        return phi

    current_t = initial_t
    nu1, nu2, nu3, nu4, nu5 = nu1_f(current_t), nu2_f(current_t), nu3_f(current_t), nu4_f(current_t), nu5_f(current_t)
    gamma1, gamma2, gamma3, gamma4, gamma5 = gamma1_f(current_t), gamma2_f(current_t), gamma3_f(current_t), gamma4_f(current_t), gamma5_f(current_t)
    h1, h2, h3, h4, h5 = h1_f(current_t), h2_f(current_t), h3_f(current_t), h4_f(current_t), h5_f(current_t)
    m12, m13, m14, m15 = m12_f(current_t), m13_f(current_t), m14_f(current_t), m15_f(current_t)
    m21, m23, m24, m25 = m21_f(current_t), m23_f(current_t), m24_f(current_t), m25_f(current_t)
    m31, m32, m34, m35 = m31_f(current_t), m32_f(current_t), m34_f(current_t), m35_f(current_t)
    m41, m42, m43, m45 = m41_f(current_t), m42_f(current_t), m43_f(current_t), m45_f(current_t)
    m51, m52, m53, m54 = m51_f(current_t), m52_f(current_t), m53_f(current_t), m54_f(current_t)

    dx,dy,dz,da,db = numpy.diff(xx),numpy.diff(yy),numpy.diff(zz),numpy.diff(aa),numpy.diff(bb)
    while current_t < T:
        dt = min(_compute_dt(dx,nu1,[m12,m13,m14,m15],gamma1,h1),
                 _compute_dt(dy,nu2,[m21,m23,m24,m25],gamma2,h2),
                 _compute_dt(dz,nu3,[m31,m32,m34,m35],gamma3,h3),
                 _compute_dt(da,nu4,[m41,m42,m43,m45],gamma4,h4),
                 _compute_dt(db,nu5,[m51,m52,m53,m54],gamma5,h5))
        this_dt = min(dt, T - current_t)

        next_t = current_t + this_dt

        nu1, nu2, nu3, nu4, nu5 = nu1_f(next_t), nu2_f(next_t), nu3_f(next_t), nu4_f(next_t), nu5_f(next_t)
        gamma1, gamma2, gamma3, gamma4, gamma5 = gamma1_f(next_t), gamma2_f(next_t), gamma3_f(next_t), gamma4_f(next_t), gamma5_f(next_t)
        h1, h2, h3, h4, h5 = h1_f(next_t), h2_f(next_t), h3_f(next_t), h4_f(next_t), h5_f(next_t)
        m12, m13, m14, m15 = m12_f(next_t), m13_f(next_t), m14_f(next_t), m15_f(next_t)
        m21, m23, m24, m25 = m21_f(next_t), m23_f(next_t), m24_f(next_t), m25_f(next_t)
        m31, m32, m34, m35 = m31_f(next_t), m32_f(next_t), m34_f(next_t), m35_f(next_t)
        m41, m42, m43, m45 = m41_f(next_t), m42_f(next_t), m43_f(next_t), m45_f(next_t)
        m51, m52, m53, m54 = m51_f(next_t), m52_f(next_t), m53_f(next_t), m54_f(next_t)
        theta0 = theta0_f(next_t)

        if numpy.any(numpy.less([T,nu1,nu2,nu3,nu4,nu5,m12,m13,m14,m15,m21,
                                 m23,m24,m25, m31,m32,m34,m35, m41,m42,m43,m45,
                                 m51,m52,m53,m54, theta0],
                                0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if numpy.any(numpy.equal([nu1,nu2,nu3,nu4,nu5], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        _inject_mutations_5D(phi, this_dt, xx, yy, zz, aa, bb, theta0,
                             frozen1, frozen2, frozen3, frozen4, frozen5)
        if not frozen1:
            phi = int_c.implicit_5Dx(phi, xx, yy, zz, aa, bb, nu1, m12, m13, m14, m15,
                                     gamma1, h1, this_dt, use_delj_trick)
        if not frozen2:
            phi = int_c.implicit_5Dy(phi, xx, yy, zz, aa, bb, nu2, m21, m23, m24, m25,
                                     gamma2, h2, this_dt, use_delj_trick)
        if not frozen3:
            phi = int_c.implicit_5Dz(phi, xx, yy, zz, aa, bb, nu3, m31, m32, m34, m35,
                                     gamma3, h3, this_dt, use_delj_trick)
        if not frozen4:
            phi = int_c.implicit_5Da(phi, xx, yy, zz, aa, bb, nu4, m41, m42, m43, m45,
                                     gamma4, h4, this_dt, use_delj_trick)
        if not frozen5:
            phi = int_c.implicit_5Db(phi, xx, yy, zz, aa, bb, nu5, m51, m52, m53, m54,
                                     gamma5, h5, this_dt, use_delj_trick)

        current_t = next_t
    return phi


#
# Here are the python versions of the population genetic functions.
#
def _Vfunc(x, nu, beta=1):
    return 1./nu * x*(1-x) * (beta+1.)**2/(4.*beta)
def _Mfunc1D(x, gamma, h):
    return gamma * 2*(h + (1-2*h)*x) * x*(1-x)
def _Mfunc2D(x,y, mxy, gamma, h):
    return mxy * (y-x) + gamma * 2*(h + (1-2*h)*x) * x*(1-x)
def _Mfunc3D(x,y,z, mxy,mxz, gamma, h):
    return mxy * (y-x) + mxz * (z-x) + gamma * 2*(h + (1-2*h)*x) * x*(1-x)

def _compute_dfactor(dx):
    r"""
    \Delta_j from the paper.
    """
    # Controls how we take the derivative of the flux. The values here depend
    #  on the fact that we're defining our probability integral using the
    #  trapezoid rule.
    dfactor = numpy.zeros(len(dx)+1)
    dfactor[1:-1] = 2/(dx[:-1] + dx[1:])
    dfactor[0] = 2/dx[0]
    dfactor[-1] = 2/dx[-1]
    return dfactor

def _compute_delj(dx, MInt, VInt, axis=0):
    r"""
    Chang an Cooper's \delta_j term. Typically we set this to 0.5.
    """
    # Chang and Cooper's fancy delta j trick...
    if use_delj_trick:
        # upslice will raise the dimensionality of dx and VInt to be appropriate
        # for functioning with MInt.
        upslice = [nuax for ii in range(MInt.ndim)]
        upslice [axis] = slice(None)

        wj = 2 *MInt*dx[upslice]
        epsj = numpy.exp(wj/VInt[upslice])
        delj = (-epsj*wj + epsj * VInt[upslice] - VInt[upslice])/(wj - epsj*wj)
        # These where statements filter out edge case for delj
        delj = numpy.where(numpy.isnan(delj), 0.5, delj)
        delj = numpy.where(numpy.isinf(delj), 0.5, delj)
    else:
        delj = 0.5
    return delj

def _one_pop_const_params(phi, xx, T, nu=1, gamma=0, h=0.5, theta0=1, 
                          initial_t=0, beta=1):
    """
    Integrate one population with constant parameters.

    In this case, we can precompute our a,b,c matrices for the linear system
    we need to evolve. This we can efficiently do in Python, rather than 
    relying on C. The nice thing is that the Python is much faster to debug.
    """
    if numpy.any(numpy.less([T,nu,theta0], 0)):
        raise ValueError('A time, population size, migration rate, or theta0 '
                         'is < 0. Has the model been mis-specified?')
    if numpy.any(numpy.equal([nu], 0)):
        raise ValueError('A population size is 0. Has the model been '
                         'mis-specified?')

    M = _Mfunc1D(xx, gamma, h)
    MInt = _Mfunc1D((xx[:-1] + xx[1:])/2, gamma, h)
    V = _Vfunc(xx, nu, beta=beta)
    VInt = _Vfunc((xx[:-1] + xx[1:])/2, nu, beta=beta)

    dx = numpy.diff(xx)
    dfactor = _compute_dfactor(dx)
    delj = _compute_delj(dx, MInt, VInt)

    a = numpy.zeros(phi.shape)
    a[1:] += dfactor[1:]*(-MInt * delj - V[:-1]/(2*dx))

    c = numpy.zeros(phi.shape)
    c[:-1] += -dfactor[:-1]*(-MInt * (1-delj) + V[1:]/(2*dx))

    b = numpy.zeros(phi.shape)
    b[:-1] += -dfactor[:-1]*(-MInt * delj - V[:-1]/(2*dx))
    b[1:] += dfactor[1:]*(-MInt * (1-delj) + V[1:]/(2*dx))

    if(M[0] <= 0):
        b[0] += (0.5/nu - M[0])*2/dx[0]
    if(M[-1] >= 0):
        b[-1] += -(-0.5/nu - M[-1])*2/dx[-1]

    dt = _compute_dt(dx,nu,[0],gamma,h)
    current_t = initial_t
    while current_t < T:    
        this_dt = min(dt, T - current_t)

        _inject_mutations_1D(phi, this_dt, xx, theta0)
        r = phi/this_dt
        phi = tridiag.tridiag(a, b+1/this_dt, c, r)
        current_t += this_dt
    return phi

def _two_pops_const_params(phi, xx, T, nu1=1,nu2=1, m12=0, m21=0,
                           gamma1=0, gamma2=0, h1=0.5, h2=0.5, theta0=1, 
                           initial_t=0, frozen1=False, frozen2=False,
                           nomut1=False, nomut2=False):
    """
    Integrate two populations with constant parameters.
    """
    if numpy.any(numpy.less([T,nu1,nu2,m12,m21,theta0], 0)):
        raise ValueError('A time, population size, migration rate, or theta0 '
                         'is < 0. Has the model been mis-specified?')
    if numpy.any(numpy.equal([nu1,nu2], 0)):
        raise ValueError('A population size is 0. Has the model been '
                         'mis-specified?')
    yy = xx

    # The use of nuax (= numpy.newaxis) here is for memory conservation. We
    # could just create big X and Y arrays which only varied along one axis,
    # but that would be wasteful.
    Vx = _Vfunc(xx, nu1)
    VxInt = _Vfunc((xx[:-1]+xx[1:])/2, nu1)
    Mx = _Mfunc2D(xx[:,nuax], yy[nuax,:], m12, gamma1, h1)
    MxInt = _Mfunc2D((xx[:-1,nuax]+xx[1:,nuax])/2, yy[nuax,:], m12, gamma1, h1)

    Vy = _Vfunc(yy, nu2)
    VyInt = _Vfunc((yy[1:]+yy[:-1])/2, nu2)
    My = _Mfunc2D(yy[nuax,:], xx[:,nuax], m21, gamma2, h2)
    MyInt = _Mfunc2D((yy[nuax,1:] + yy[nuax,:-1])/2, xx[:,nuax], m21, gamma2,h2)

    dx = numpy.diff(xx)
    dfact_x = _compute_dfactor(dx)
    deljx = _compute_delj(dx, MxInt, VxInt)

    dy = numpy.diff(yy)
    dfact_y = _compute_dfactor(dy)
    deljy = _compute_delj(dy, MyInt, VyInt, axis=1)

    # The nuax's here broadcast the our various arrays to have the proper shape
    # to fit into ax,bx,cx
    ax, bx, cx = [numpy.zeros(phi.shape) for ii in range(3)]
    ax[ 1:] += dfact_x[ 1:,nuax]*(-MxInt*deljx    - Vx[:-1,nuax]/(2*dx[:,nuax]))
    cx[:-1] += dfact_x[:-1,nuax]*( MxInt*(1-deljx)- Vx[ 1:,nuax]/(2*dx[:,nuax]))
    bx[:-1] += dfact_x[:-1,nuax]*( MxInt*deljx    + Vx[:-1,nuax]/(2*dx[:,nuax]))
    bx[ 1:] += dfact_x[ 1:,nuax]*(-MxInt*(1-deljx)+ Vx[ 1:,nuax]/(2*dx[:,nuax]))

    if Mx[0,0] <= 0:
        bx[0,0] += (0.5/nu1 - Mx[0,0])*2/dx[0]
    if Mx[-1,-1] >= 0:
        bx[-1,-1] += -(-0.5/nu1 - Mx[-1,-1])*2/dx[-1]

    ay, by, cy = [numpy.zeros(phi.shape) for ii in range(3)]
    ay[:, 1:] += dfact_y[ 1:]*(-MyInt*deljy     - Vy[nuax,:-1]/(2*dy))
    cy[:,:-1] += dfact_y[:-1]*( MyInt*(1-deljy) - Vy[nuax, 1:]/(2*dy))
    by[:,:-1] += dfact_y[:-1]*( MyInt*deljy     + Vy[nuax,:-1]/(2*dy))
    by[:, 1:] += dfact_y[ 1:]*(-MyInt*(1-deljy) + Vy[nuax, 1:]/(2*dy))

    if My[0,0] <= 0:
        by[0,0] += (0.5/nu2 - My[0,0])*2/dy[0]
    if My[-1,-1] >= 0:
        by[-1,-1] += -(-0.5/nu2 - My[-1,-1])*2/dy[-1]

    dt = min(_compute_dt(dx,nu1,[m12],gamma1,h1),
             _compute_dt(dy,nu2,[m21],gamma2,h2))
    current_t = initial_t

    if cuda_enabled:
        import dadi.cuda
        phi = dadi.cuda.Integration._two_pops_const_params(phi, xx,
                theta0, frozen1, frozen2, nomut1, nomut2, ax, bx, cx, ay,
                by, cy, current_t, dt, T)
        return phi

    while current_t < T:
        this_dt = min(dt, T - current_t)
        _inject_mutations_2D(phi, this_dt, xx, yy, theta0, frozen1, frozen2,
                            nomut1, nomut2)
        if not frozen1:
            phi = int_c.implicit_precalc_2Dx(phi, ax, bx, cx, this_dt)
        if not frozen2:
            phi = int_c.implicit_precalc_2Dy(phi, ay, by, cy, this_dt)
        current_t += this_dt

    return phi

def _three_pops_const_params(phi, xx, T, nu1=1, nu2=1, nu3=1, 
                             m12=0, m13=0, m21=0, m23=0, m31=0, m32=0, 
                             gamma1=0, gamma2=0, gamma3=0, 
                             h1=0.5, h2=0.5, h3=0.5, theta0=1, initial_t=0,
                             frozen1=False, frozen2=False, frozen3=False):
    """
    Integrate three population with constant parameters.
    """
    if numpy.any(numpy.less([T,nu1,nu2,nu3,m12,m13,m21,m23,m31,m32,theta0], 0)):
        raise ValueError('A time, population size, migration rate, or theta0 '
                         'is < 0. Has the model been mis-specified?')
    if numpy.any(numpy.equal([nu1,nu2,nu3], 0)):
        raise ValueError('A population size is 0. Has the model been '
                         'mis-specified?')
    zz = yy = xx

    Vx = _Vfunc(xx, nu1)
    VxInt = _Vfunc((xx[:-1]+xx[1:])/2, nu1)
    Mx = _Mfunc3D(xx[:,nuax,nuax], yy[nuax,:,nuax], zz[nuax,nuax,:], 
                 m12, m13, gamma1, h1)
    MxInt = _Mfunc3D((xx[:-1,nuax,nuax]+xx[1:,nuax,nuax])/2, yy[nuax,:,nuax], 
                    zz[nuax,nuax,:], m12, m13, gamma1, h1)

    dx = numpy.diff(xx)
    dfact_x = _compute_dfactor(dx)
    deljx = _compute_delj(dx, MxInt, VxInt)

    ax, bx, cx = [numpy.zeros(phi.shape) for ii in range(3)]
    ax[ 1:] += dfact_x[ 1:,nuax,nuax]*(-MxInt*deljx    
                                       - Vx[:-1,nuax,nuax]/(2*dx[:,nuax,nuax]))
    cx[:-1] += dfact_x[:-1,nuax,nuax]*( MxInt*(1-deljx)
                                       - Vx[ 1:,nuax,nuax]/(2*dx[:,nuax,nuax]))
    bx[:-1] += dfact_x[:-1,nuax,nuax]*( MxInt*deljx    
                                       + Vx[:-1,nuax,nuax]/(2*dx[:,nuax,nuax]))
    bx[ 1:] += dfact_x[ 1:,nuax,nuax]*(-MxInt*(1-deljx)
                                       + Vx[ 1:,nuax,nuax]/(2*dx[:,nuax,nuax]))

    if Mx[0,0,0] <= 0:
        bx[0,0,0] += (0.5/nu1 - Mx[0,0,0])*2/dx[0]
    if Mx[-1,-1,-1] >= 0:
        bx[-1,-1,-1] += -(-0.5/nu1 - Mx[-1,-1,-1])*2/dx[-1]

    # Memory consumption can be an issue in 3D, so we delete arrays after we're
    # done with them.
    del Vx,VxInt,Mx,MxInt,deljx

    Vy = _Vfunc(yy, nu2)
    VyInt = _Vfunc((yy[1:]+yy[:-1])/2, nu2)
    My = _Mfunc3D(yy[nuax,:,nuax], xx[:,nuax, nuax], zz[nuax,nuax,:],
                 m21, m23, gamma2, h2)
    MyInt = _Mfunc3D((yy[nuax,1:,nuax] + yy[nuax,:-1,nuax])/2, xx[:,nuax, nuax],
                    zz[nuax,nuax,:], m21, m23, gamma2, h2)

    dy = numpy.diff(yy)
    dfact_y = _compute_dfactor(dy)
    deljy = _compute_delj(dy, MyInt, VyInt, axis=1)

    ay, by, cy = [numpy.zeros(phi.shape) for ii in range(3)]
    ay[:, 1:] += dfact_y[nuax, 1:,nuax]*(-MyInt*deljy     
                                    - Vy[nuax,:-1,nuax]/(2*dy[nuax,:,nuax]))
    cy[:,:-1] += dfact_y[nuax,:-1,nuax]*( MyInt*(1-deljy) 
                                    - Vy[nuax, 1:,nuax]/(2*dy[nuax,:,nuax]))
    by[:,:-1] += dfact_y[nuax,:-1,nuax]*( MyInt*deljy     
                                    + Vy[nuax,:-1,nuax]/(2*dy[nuax,:,nuax]))
    by[:, 1:] += dfact_y[nuax, 1:,nuax]*(-MyInt*(1-deljy) 
                                    + Vy[nuax, 1:,nuax]/(2*dy[nuax,:,nuax]))

    if My[0,0,0] <= 0:
        by[0,0,0] += (0.5/nu2 - My[0,0,0])*2/dy[0]
    if My[-1,-1,-1] >= 0:
        by[-1,-1,-1] += -(-0.5/nu2 - My[-1,-1,-1])*2/dy[-1]

    del Vy,VyInt,My,MyInt,deljy

    Vz = _Vfunc(zz, nu3)
    VzInt = _Vfunc((zz[1:]+zz[:-1])/2, nu3)
    Mz = _Mfunc3D(zz[nuax,nuax,:], xx[:,nuax, nuax], yy[nuax,:,nuax],
                 m31, m32, gamma3, h3)
    MzInt = _Mfunc3D((zz[nuax,nuax,1:] + zz[nuax,nuax,:-1])/2, xx[:,nuax, nuax],
                    yy[nuax,:,nuax], m31, m32, gamma3, h3)

    dz = numpy.diff(zz)
    dfact_z = _compute_dfactor(dz)
    deljz = _compute_delj(dz, MzInt, VzInt, axis=2)

    az, bz, cz = [numpy.zeros(phi.shape) for ii in range(3)]
    az[:,:, 1:] += dfact_z[ 1:]*(-MzInt*deljz     - Vz[nuax,nuax,:-1]/(2*dz))
    cz[:,:,:-1] += dfact_z[:-1]*( MzInt*(1-deljz) - Vz[nuax,nuax, 1:]/(2*dz))
    bz[:,:,:-1] += dfact_z[:-1]*( MzInt*deljz     + Vz[nuax,nuax,:-1]/(2*dz))
    bz[:,:, 1:] += dfact_z[ 1:]*(-MzInt*(1-deljz) + Vz[nuax,nuax, 1:]/(2*dz))

    if Mz[0,0,0] <= 0:
        bz[0,0,0] += (0.5/nu3 - Mz[0,0,0])*2/dz[0]
    if Mz[-1,-1,-1] >= 0:
        bz[-1,-1,-1] += -(-0.5/nu3 - Mz[-1,-1,-1])*2/dz[-1]

    del Vz,VzInt,Mz,MzInt,deljz

    dt = min(_compute_dt(dx,nu1,[m12,m13],gamma1,h1),
             _compute_dt(dy,nu2,[m21,m23],gamma2,h2),
             _compute_dt(dz,nu3,[m31,m32],gamma3,h3))
    current_t = initial_t

    if cuda_enabled:
        import dadi.cuda
        phi = dadi.cuda.Integration._three_pops_const_params(phi, xx,
                theta0, frozen1, frozen2, frozen3, 
                ax, bx, cx, ay, by, cy, az, bz, cz,
                current_t, dt, T)
        return phi

    while current_t < T:    
        this_dt = min(dt, T - current_t)
        _inject_mutations_3D(phi, this_dt, xx, yy, zz, theta0,
                             frozen1, frozen2, frozen3)
        if not frozen1:
            phi = int_c.implicit_precalc_3Dx(phi, ax, bx, cx, this_dt)
        if not frozen2:
            phi = int_c.implicit_precalc_3Dy(phi, ay, by, cy, this_dt)
        if not frozen3:
            phi = int_c.implicit_precalc_3Dz(phi, az, bz, cz, this_dt)
        current_t += this_dt
    return phi

def _Vfunc_X(x, nu, beta):
    return 1./nu * x*(1-x) * (2*beta+4.)*(beta+1.)/(9.*beta)
def _Mfunc1D_X(x, gamma, h, beta):
    return gamma * 4./3. * (0.5+h+x*(1-2*h)) * x*(1-x)
def _inject_mutations_1D_X(phi, dt, xx, theta0, beta, alpha):
    """
    Inject novel mutations for a timestep.
    """
    factor = 2./(1.+2.*beta)*(1./(alpha+1.) + beta) 
    phi[1] += dt/xx[1] * theta0/2. * factor * 2./(xx[2] - xx[0])
    return phi

def one_pop_X(phi, xx, T, nu=1, gamma=0, h=0.5, beta=1, alpha=1, theta0=1.0, 
              initial_t=0, frozen=False):
    """
    Integrate a 1-dimensional phi foward.

    phi: Initial 1-dimensional phi
    xx: Grid upon (0,1) overwhich phi is defined.

    nu, gamma, and theta0 may be functions of time.
    nu: Population size
    gamma: Scaled selection coefficient on *all* segregating alleles
    h: Dominance coefficient. h = 0.5 corresponds to genic selection. 
       Heterozygous females have fitness 1+2sh and homozygous females have
       fitness 1+2s. Male carriers have fitness 1+2s.
    theta0: Propotional to ancestral size. Typically constant.
    beta: Breeding ratio, beta=Nf/Nm.
    alpha: Male to female mutation rate ratio, beta = mu_m / mu_f.

    T: Time at which to halt integration
    initial_t: Time at which to start integration. (Note that this only matters
               if one of the demographic parameters is a function of time.)

    frozen: If True, population is 'frozen' so that it does not change.
            In the one_pop case, this is equivalent to not running the
            integration at all.
    """
    phi = phi.copy()

    # For a one population integration, freezing means just not integrating.
    if frozen:
        return phi

    if T - initial_t == 0:
        return phi
    elif T - initial_t < 0:
        raise ValueError('Final integration time T (%f) is less than '
                         'intial_time (%f). Integration cannot be run '
                         'backwards.' % (T, initial_t))

    vars_to_check = (nu, gamma, h, theta0, beta, alpha)
    if numpy.all([numpy.isscalar(var) for var in vars_to_check]):
        return _one_pop_const_params_X(phi, xx, T, nu, gamma, h, beta, alpha, 
                                       theta0, initial_t)
    else:
        raise NotImplementedError('X chromosome integration currently only '
                                  'implemented for constant parameters.')

def _one_pop_const_params_X(phi, xx, T, nu=1, gamma=0, h=0.5, beta=1, alpha=1, 
                            theta0=1, initial_t=0):
    """
    Integrate one population with constant parameters.

    In this case, we can precompute our a,b,c matrices for the linear system
    we need to evolve. This we can efficiently do in Python, rather than 
    relying on C. The nice thing is that the Python is much faster to debug.
    """
    if numpy.any(numpy.less([T,nu,theta0,beta,alpha], 0)):
        raise ValueError('A time, population size, migration rate, theta0, '
                         'beta, or alpha is < 0. Has the model been '
                         'mis-specified?')
    if numpy.any(numpy.equal([nu,beta], 0)):
        raise ValueError('A population size or beta is 0. Has the model been '
                         'mis-specified?')

    M = _Mfunc1D_X(xx, gamma, h, beta)
    MInt = _Mfunc1D_X((xx[:-1] + xx[1:])/2, gamma, h, beta)
    V = _Vfunc_X(xx, nu, beta)
    VInt = _Vfunc_X((xx[:-1] + xx[1:])/2, nu, beta)

    dx = numpy.diff(xx)
    dfactor = _compute_dfactor(dx)
    delj = _compute_delj(dx, MInt, VInt)

    a = numpy.zeros(phi.shape)
    a[1:] += dfactor[1:]*(-MInt * delj - V[:-1]/(2*dx))

    c = numpy.zeros(phi.shape)
    c[:-1] += -dfactor[:-1]*(-MInt * (1-delj) + V[1:]/(2*dx))

    b = numpy.zeros(phi.shape)
    b[:-1] += -dfactor[:-1]*(-MInt * delj - V[:-1]/(2*dx))
    b[1:] += dfactor[1:]*(-MInt * (1-delj) + V[1:]/(2*dx))

    if(M[0] <= 0):
        b[0] += (0.5/nu - M[0])*2/dx[0]
    if(M[-1] >= 0):
        b[-1] += -(-0.5/nu - M[-1])*2/dx[-1]

    dt = _compute_dt(dx,nu,[0],gamma,h)
    current_t = initial_t
    while current_t < T:    
        this_dt = min(dt, T - current_t)

        _inject_mutations_1D_X(phi, this_dt, xx, theta0, beta, alpha)
        r = phi/this_dt
        phi = tridiag.tridiag(a, b+1./this_dt, c, r)
        current_t += this_dt
    return phi



"""
Miscellaneous functions for plotting DFEs
"""
import numpy as np

def plot_biv_dfe(gammax, gammay, sel_dist, params, logweight=True, ax=None,
                 xlabel='$\gamma_1$', ylabel='$\gamma_2$', cmap='gray_r',
                 colorbar=True, vmin=0, clip_on=True):
    """
    Plot bivariate DFE (for negative gammas).

    gammax, gammay: Grids of gamma values to plot with. Note that non-negative
                    values are discarded.
    sel_dist: Bivariate probability distribution,
              taking in arguments (xx, yy, params)
    params: Parameters for sel_dist
    logweight: If True, plotted values are weighted by x and y, so that the
               total probability within each cell is plotted, rather than the
               probability density. This is typically easier to interpret.
    ax: Matplotlib axes to plot into. If None, plt.gca() is used.
    xlabel, ylabel: Labels for x and y axes.
    cmap: Colormap to plot with. For useful advice regarding colormaps, see
          http://matplotlib.org/users/colormaps.html .
    colorbar: If True, include a colorbar alongside the plot.
    vmin: Values below this will be colored white.
    clip_on: If False, plot will extend outside of axes.
    """
    # Discard non-negative values of gamma.
    gammax = gammax[gammax < 0]
    gammay = gammay[gammay < 0]

    # Midpoints of each gamma bin (on a log-scale)
    xmid = np.sqrt(gammax[:-1] * gammax[1:])
    ymid = np.sqrt(gammay[:-1] * gammay[1:])
    # Calculate weights
    weights = sel_dist(xmid, ymid, params)
    # Convert from probability density to probability value within each cell.
    if logweight:
        weights *= xmid[:,np.newaxis] * ymid[np.newaxis,:]

    # Plot data
    X,Y = np.meshgrid(gammax, gammay)

    if not ax:
        # Hiding import here, so Selection_2d.py isn't dependent on matplotlib.
        import matplotlib.pyplot as plt
        ax = plt.gca()
    masked = np.ma.masked_where(weights.T<vmin,weights.T)
    mappable = ax.pcolor(X,Y,masked, cmap=cmap, vmin=vmin, clip_on=clip_on)

    # Plot the colorbar, labeling carefully depending on logweight
    if colorbar:
        cb = ax.get_figure().colorbar(mappable)
        if logweight:
            cb.set_label('Probability')
        else:
            cb.set_label('Probability density')

    # Set to logscale. Use symlog because it handles negative values cleanly
    ax.set_xscale('symlog', linthreshx=abs(gammax).min())
    ax.set_yscale('symlog', linthreshy=abs(gammay).min())

    ax.set_xlabel(xlabel, fontsize='large')
    ax.set_ylabel(ylabel, fontsize='large')

    return ax

def plot_biv_point_pos_dfe(gammax, gammay, sel_dist, params, rho=0,
                           logweight=True, xlabel='$\gamma_1$',
                           ylabel='$\gamma_2$', cmap='gray_r', fignum=None,
                           colorbar=True, vmin=0):
    """
    Plot bivariate DFE (for negative gammas) with positive point mass.

    Returns figure for plot.

    Note: You might need to adjust subplot parameters using fig.subplots_adjust
          after plotting to see all labels, etc.

    gammax, gammay: Grids of gamma values to plot with. Note that non-negative
                    values are discarded.
    sel_dist: Bivariate probability distribution,
              taking in arguments (xx, yy, params)
    params: Parameters for sel_dist and positive selection.
            It is assumed that the last four parameters are:
            Proportion positive selection in pop1, postive gamma for pop1,
             prop. positive in pop2, and positive gamma for pop2.
            Earlier arguments are assumed to be for the continuous bivariate
            distribution.
    rho: Correlation coefficient used to connect negative and positive
         components of DFE.
    logweight: If True, plotted values are weighted by x and y, so that the
               total probability within each cell is plotted, rather than the
               probability density. This is typically easier to interpret.
    xlabel, ylabel: Labels for x and y axes.
    cmap: Colormap to plot with. For useful advice regarding colormaps, see
          http://matplotlib.org/users/colormaps.html .
    fignum: Figure number to use. If None or figure does not exist, a new
            one will be created.
    colorbar: If True, plot scale bar for probability
    vmin: Values below this will be colored white.
    """
    biv_params = params[:-4]
    ppos1, gammapos1, ppos2, gammapos2 = params[-4:]

    # Pull out negative gammas and calculate (log) midpoints.
    gammax = gammax[gammax < 0]
    gammay = gammay[gammay < 0]
    xmid = np.sqrt(gammax[:-1] * gammax[1:])
    ymid = np.sqrt(gammay[:-1] * gammay[1:])

    # Bivariate negative part of DFE
    neg_neg = sel_dist(xmid, ymid, biv_params)
    # Marginal DFEs from integrating along each axis
    neg_pos = np.trapz(neg_neg, ymid, axis=1)
    pos_neg = np.trapz(neg_neg, xmid, axis=0)

    if logweight:
        neg_neg *= xmid[:,np.newaxis] * ymid[np.newaxis,:]
        neg_pos *= xmid
        pos_neg *= ymid

    # Weighting factors for each quadrant of the DFE. Note that in the case
    # that ppos1==ppos2, these reduce to the model of Ragsdale et al. (2016)
    p_pos_pos = ppos1*ppos2 + rho*(np.sqrt(ppos1*ppos2) - ppos1*ppos2)
    p_pos_neg = (1-rho) * ppos1*(1-ppos2)
    p_neg_pos = (1-rho) * (1-ppos1)*ppos2
    p_neg_neg = (1-ppos1)*(1-ppos2)\
            + rho*(1-np.sqrt(ppos1*ppos2) - (1-ppos1)*(1-ppos2))

    # Apply weighting factors
    pos_neg *= p_pos_neg
    neg_pos *= p_neg_pos
    neg_neg *= p_neg_neg

    # For plotting, to put all on same color-scale.
    vmax = max([neg_neg.max(), neg_pos.max(), pos_neg.max(), p_pos_pos])

    # Grid points for plotting in positive selection regime.
    if gammapos1 != gammapos2:
        pos_grid = np.array(sorted([gammapos1, (gammapos1 + gammapos2)/2.,
                                    gammapos2]))
    else:
        pos_grid = np.array([gammapos1-1, gammapos1, gammapos1+1])

    # Fill in grids for plotting positive terms
    pos_neg_grid = np.zeros((3, len(ymid)))
    pos_neg_grid[pos_grid == gammapos1] = pos_neg
    neg_pos_grid = np.zeros((len(xmid), 3))
    neg_pos_grid[:,pos_grid == gammapos2] = neg_pos[:,np.newaxis]
    pos_pos_grid = np.zeros((3,3))
    pos_pos_grid[pos_grid == gammapos1, pos_grid == gammapos2] = p_pos_pos

    # For plotting with pcolor, need grid one size bigger.
    d = pos_grid[1] - pos_grid[0]
    plot_pos_grid = np.linspace(pos_grid[0] - d/2., pos_grid[-1] + d/2., 4)

    import matplotlib.pyplot as plt
    fig = plt.figure(num=fignum, figsize=(4,3), dpi=150)
    fig.clear()

    gs = plt.matplotlib.gridspec.GridSpec(2,3, width_ratios=[9,1,0.5],
                                          height_ratios=[1,9],
                                          wspace=0.15, hspace=0.15)
    # Create our four axes
    ax_ll = fig.add_subplot(gs[1,0])
    ax_ul = fig.add_subplot(gs[0,0], sharex=ax_ll)
    ax_lr = fig.add_subplot(gs[1,1], sharey=ax_ll)
    ax_ur = fig.add_subplot(gs[0,1], sharex=ax_lr, sharey=ax_ul)

    # Plot in each axis
    X,Y = np.meshgrid(gammax, gammay)
    masked = np.ma.masked_where(neg_neg.T<vmin,neg_neg.T)
    mappable = ax_ll.pcolor(X,Y,masked, cmap=cmap, vmin=vmin, vmax=vmax)
    X,Y = np.meshgrid(plot_pos_grid, plot_pos_grid)
    masked = np.ma.masked_where(pos_pos_grid.T<vmin,pos_pos_grid.T)
    ax_ur.pcolor(X,Y,masked, cmap=cmap, vmin=vmin, vmax=vmax)
    X,Y = np.meshgrid(plot_pos_grid, gammay)
    masked = np.ma.masked_where(pos_neg_grid.T<vmin,pos_neg_grid.T)
    ax_lr.pcolor(X,Y,masked, cmap=cmap, vmin=vmin, vmax=vmax)
    X,Y = np.meshgrid(gammax, plot_pos_grid)
    masked = np.ma.masked_where(neg_pos_grid.T<vmin,neg_pos_grid.T)
    ax_ul.pcolor(X,Y,masked, cmap=cmap, vmin=vmin, vmax=vmax)

    # Set logarithmic scale on legative parts
    ax_ll.set_xscale('symlog', linthreshx=abs(gammax).min())
    ax_ll.set_yscale('symlog', linthreshy=abs(gammay).min())

    # Remove interior axes lines and tickmarks
    ax_ll.spines['right'].set_visible(False)
    ax_ll.spines['top'].set_visible(False)
    ax_ll.tick_params('both', top=False, right=False, which='both')

    ax_ul.spines['right'].set_visible(False)
    ax_ul.spines['bottom'].set_visible(False)
    ax_ul.tick_params('both', top=True, bottom=False,
                      labelbottom=False, right=False, which='both')

    ax_lr.spines['left'].set_visible(False)
    ax_lr.spines['top'].set_visible(False)
    ax_lr.tick_params('both', right=True, left=False,
                      labelleft=False, top=False, which='both')

    ax_ur.spines['left'].set_visible(False)
    ax_ur.spines['bottom'].set_visible(False)
    ax_ur.tick_params('both', right=True, left=False,
                      labelleft=False, top=True, bottom=False,
                      labelbottom=False, which='both')

    # Set tickmarks for positive selection
    if gammapos1 != gammapos2:
        ax_ur.set_xticks([gammapos1, gammapos2])
        ax_ur.set_yticks([gammapos1, gammapos2])
    else:
        ax_ur.set_xticks([gammapos1])
        ax_ur.set_yticks([gammapos1])

    ax_ll.set_xlabel(xlabel, fontsize='large')
    ax_ll.set_ylabel(ylabel, fontsize='large')

    if colorbar:
        ax = fig.add_subplot(gs[1,2])
        cb = fig.colorbar(mappable, cax=ax, use_gridspec=True)
        if logweight:
            cb.set_label('Probability')
    fig.subplots_adjust(top=0.99, right=0.82, bottom=0.19, left=0.18)

    return fig


"""
Single population demographic models.
"""
import numpy

from dadi import Numerics, PhiManip, Integration
from dadi.Spectrum_mod import Spectrum

def snm(notused, ns, pts):
    """
    Standard neutral model.

    ns = (n1,)

    n1: Number of samples in resulting Spectrum
    pts: Number of grid points to use in integration.
    """
    xx = Numerics.default_grid(pts)
    phi = PhiManip.phi_1D(xx)

    fs = Spectrum.from_phi(phi, ns, (xx,))
    return fs

def two_epoch(params, ns, pts):
    """
    Instantaneous size change some time ago.

    params = (nu,T)
    ns = (n1,)

    nu: Ratio of contemporary to ancient population size
    T: Time in the past at which size change happened (in units of 2*Na 
       generations) 
    n1: Number of samples in resulting Spectrum
    pts: Number of grid points to use in integration.
    """
    nu,T = params

    xx = Numerics.default_grid(pts)
    phi = PhiManip.phi_1D(xx)
    
    phi = Integration.one_pop(phi, xx, T, nu)

    fs = Spectrum.from_phi(phi, ns, (xx,))
    return fs

def growth(params, ns, pts):
    """
    Exponential growth beginning some time ago.

    params = (nu,T)
    ns = (n1,)

    nu: Ratio of contemporary to ancient population size
    T: Time in the past at which growth began (in units of 2*Na 
       generations) 
    n1: Number of samples in resulting Spectrum
    pts: Number of grid points to use in integration.
    """
    nu,T = params

    xx = Numerics.default_grid(pts)
    phi = PhiManip.phi_1D(xx)

    nu_func = lambda t: numpy.exp(numpy.log(nu) * t/T)
    phi = Integration.one_pop(phi, xx, T, nu_func)

    fs = Spectrum.from_phi(phi, ns, (xx,))
    return fs

def bottlegrowth(params, ns, pts):
    """
    Instantanous size change followed by exponential growth.

    params = (nuB,nuF,T)
    ns = (n1,)

    nuB: Ratio of population size after instantanous change to ancient
         population size
    nuF: Ratio of contemporary to ancient population size
    T: Time in the past at which instantaneous change happened and growth began
       (in units of 2*Na generations) 
    n1: Number of samples in resulting Spectrum
    pts: Number of grid points to use in integration.
    """
    nuB,nuF,T = params

    xx = Numerics.default_grid(pts)
    phi = PhiManip.phi_1D(xx)

    nu_func = lambda t: nuB*numpy.exp(numpy.log(nuF/nuB) * t/T)
    phi = Integration.one_pop(phi, xx, T, nu_func)

    fs = Spectrum.from_phi(phi, ns, (xx,))
    return fs

def three_epoch(params, ns, pts):
    """
    params = (nuB,nuF,TB,TF)
    ns = (n1,)

    nuB: Ratio of bottleneck population size to ancient pop size
    nuF: Ratio of contemporary to ancient pop size
    TB: Length of bottleneck (in units of 2*Na generations) 
    TF: Time since bottleneck recovery (in units of 2*Na generations) 

    n1: Number of samples in resulting Spectrum
    pts: Number of grid points to use in integration.
    """
    nuB,nuF,TB,TF = params

    xx = Numerics.default_grid(pts)
    phi = PhiManip.phi_1D(xx)

    phi = Integration.one_pop(phi, xx, TB, nuB)
    phi = Integration.one_pop(phi, xx, TF, nuF)

    fs = Spectrum.from_phi(phi, ns, (xx,))
    return fs



"""
Probability density functions for defining DFEs.
"""
import numpy as np
import scipy.stats.distributions as ssd

def exponential(xx, params):
    """
    params: [scale]
    """
    return ssd.expon.pdf(xx, scale=params[0])

def lognormal(xx, params):
    """
    params: [mu, sigma] (in log scale)
    """
    mu, sigma = params
    return ssd.lognorm.pdf(xx, sigma, scale=np.exp(mu))

def gamma(xx, params):
    """
    params: [alpha, beta] = [shape, scale]
    """
    alpha, beta = params
    return ssd.gamma.pdf(xx, alpha, scale=beta)

def beta(xx, params):
    """
    params: [alpha, beta]
    """
    alpha, beta = params
    return ssd.beta.pdf(xx, alpha, beta)

def normal(xx, mu, sigma):
    """
    params: [mu, sigma]
    """
    return ssd.norm.pdf(xx, loc=mu, scale=sigma)

from dadi.DFE.PDFs_c import biv_lognormal, biv_ind_gamma
# Note: This method has been deprecated in favor of the much faster C version
# defined in PDFS_c
def biv_lognormal_py(xx, yy, params):
    """
    Bivariate lognormal pdf

    xx: x-coordinates at which to evaluate.
    yy: y-coordinates at which to evaluate.
    params: Input parameters. If len(params) == 3, then params = (mu,sigma,rho)
            and mu and sigma are assumed to be equal in the two dimensions. If
            len(params) == 5, then params = (mu1,mu2,sigma1,sigma2,rho)
    """
    # The atleast_1d calls here and the squeeze at the end enable this to work
    # for scalar and array arguments.
    xx = np.atleast_1d(xx)
    yy = np.atleast_1d(yy)
    if len(params) == 3:
        mu, sigma, rho = params
        mu1 = mu2 = mu
        sigma1 = sigma2 = sigma
    elif len(params) == 5:
        mu1, mu2, sigma1, sigma2, rho = params
    else:
        raise ValueError('Parameter array for bivariate lognormal must have '
                         'length 3 or 5.')
    delx = (np.log(xx[:,np.newaxis]) - mu1)/sigma1
    dely = (np.log(yy[np.newaxis,:]) - mu2)/sigma2
    norm = 2*np.pi * sigma1*sigma2 * np.sqrt(1.-rho**2) * np.outer(xx,yy)
    q = (delx**2 - 2.*rho*delx*dely + dely**2)/(1.-rho**2)

    return np.squeeze(np.exp(-q/2.)/norm)

# Note: This method has been deprecated in favor of the much faster C version
# defined in PDFS_c
def biv_ind_gamma_py(xx, yy, params):
    """
    Bivariate independent gamma pdf

    xx: x-coordinates at which to evaluate.
    yy: y-coordinates at which to evaluate.
    params: Input parameters. If len(params) == 2, then params = (alpha,beta)
            and alpha and beta are assumed to be equal in the two dimensions. If
            len(params) == 4, then params = (alpha1,alpha2,beta1,beta2).
            If len(params) in [3,5], then the last parameter is ignored (for
            compatibility with mixture model functions).

    For extensions to correlated gamma distributions, see
    Kibble (1941) and Smith and Adelfang (1981).
    """
    xx = np.atleast_1d(xx)
    yy = np.atleast_1d(yy)
    if len(params) in [2,3]:
        alpha1 = alpha2 = params[0]
        beta1 = beta2 = params[1]
    elif len(params) in [4,5]:
        alpha1, alpha2, beta1, beta2 = params[:4]
    else:
        raise ValueError('Parameter array for bivariate independent gamma must have '
                         'length 2,3,4, or 5.')

    xmarg = ssd.gamma.pdf(xx, alpha1, scale=beta1)
    ymarg = ssd.gamma.pdf(yy, alpha2, scale=beta2)
    return np.squeeze(np.outer(xmarg, ymarg))


import numpy as np

def _inject_mutations_2D_valcalc(dt, xx, yy, theta0, frozen1, frozen2,
                                 nomut1, nomut2):
    """
    Calculate mutations that need to be injected.
    """
    val10, val01 = 0, 0
    # Population 1
    if not frozen1 and not nomut1:
        val10 = dt/xx[1] * theta0/2 * 4/((xx[2] - xx[0]) * yy[1])
    # Population 2
    if not frozen2 and not nomut2:
        val01 = dt/yy[1] * theta0/2 * 4/((yy[2] - yy[0]) * xx[1])
    return np.float64(val10),np.float64(val01)

def _inject_mutations_3D_valcalc(dt, xx, yy, zz, theta0, frozen1, frozen2, frozen3):
    """
    Calculate mutations that need to be injected.
    """
    # Population 1
    # Normalization based on the multi-dimensional trapezoid rule is 
    # implemented                      ************** here ***************
    val100, val010, val001 = 0, 0, 0
    if not frozen1:
        val100 = dt/xx[1] * theta0/2 * 8/((xx[2] - xx[0]) * yy[1] * zz[1])
    # Population 2
    if not frozen2:
        val010 = dt/yy[1] * theta0/2 * 8/((yy[2] - yy[0]) * xx[1] * zz[1])
    # Population 3
    if not frozen3:
        val001 = dt/zz[1] * theta0/2 * 8/((zz[2] - zz[0]) * xx[1] * yy[1])
    return np.float64(val100), np.float64(val010), np.float64(val001)

def _inject_mutations_4D_valcalc(dt, xx, yy, zz, aa, theta0, frozen1, frozen2, frozen3, frozen4):
    """
    Calculate mutations that need to be injected.
    """
    # Population 1
    # Normalization based on the multi-dimensional trapezoid rule is 
    # implemented                      ************** here ***************
    val1000, val0100, val0010, val0001 = 0, 0, 0, 0
    if not frozen1:
        val1000 = dt/xx[1] * theta0/2 * 16/((xx[2] - xx[0]) * yy[1] * zz[1] * aa[1])
    # Population 2
    if not frozen2:
        val0100 = dt/yy[1] * theta0/2 * 16/((yy[2] - yy[0]) * xx[1] * zz[1] * aa[1])
    # Population 3
    if not frozen3:
        val0010 = dt/zz[1] * theta0/2 * 16/((zz[2] - zz[0]) * xx[1] * yy[1] * aa[1])
    # Population 4
    if not frozen4:
        val0001 = dt/aa[1] * theta0/2 * 16/((aa[2] - aa[0]) * xx[1] * yy[1] * zz[1])
    return np.float64(val1000), np.float64(val0100), np.float64(val0010), np.float64(val0001)

def _inject_mutations_5D_valcalc(dt, xx, yy, zz, aa, bb, theta0, frozen1, frozen2, frozen3, frozen4, frozen5):
    """
    Calculate mutations that need to be injected.
    """
    # Population 1
    # Normalization based on the multi-dimensional trapezoid rule is 
    # implemented                      ************** here ***************
    val10000, val01000, val00100, val00010, val00001 = 0, 0, 0, 0, 0
    if not frozen1:
        val10000 = dt/xx[1] * theta0/2 * 32/((xx[2] - xx[0]) * yy[1] * zz[1] * aa[1] * bb[1])
    # Population 2
    if not frozen2:
        val01000 = dt/yy[1] * theta0/2 * 32/((yy[2] - yy[0]) * xx[1] * zz[1] * aa[1] * bb[1])
    # Population 3
    if not frozen3:
        val00100 = dt/zz[1] * theta0/2 * 32/((zz[2] - zz[0]) * xx[1] * yy[1] * aa[1] * bb[1])
    # Population 4
    if not frozen4:
        val00010 = dt/aa[1] * theta0/2 * 32/((aa[2] - aa[0]) * xx[1] * yy[1] * zz[1] * bb[1])
    if not frozen5:
        val00001 = dt/bb[1] * theta0/2 * 32/((aa[2] - aa[0]) * xx[1] * yy[1] * zz[1] * aa[1])
    return np.float64(val10000), np.float64(val01000), np.float64(val00100), np.float64(val00010), np.float64(val00001)

import pycuda
import pycuda.gpuarray as gpuarray

import dadi.cuda
from dadi.cuda import cusparse_handle, _grid, _block, transpose_gpuarray
from dadi.cuda.cusparse import cusparseDgtsvInterleavedBatch_bufferSizeExt, cusparseDgtsvInterleavedBatch
from . import kernels

def _two_pops_const_params(phi, xx, 
                theta0, frozen1, frozen2, nomut1, nomut2,
                ax, bx, cx, ay, by, cy, 
                current_t, dt, T):
    L = np.int32(len(xx))

    ax_gpu = gpuarray.to_gpu(ax)
    bx_gpu = gpuarray.to_gpu(bx)
    ay_gpu = gpuarray.to_gpu(ay.transpose().copy())
    by_gpu = gpuarray.to_gpu(by.transpose().copy())

    # du is modified when InterleavedBatch runs.
    # Save a copy on the GPU to refill after each call.
    cx_saved_gpu = gpuarray.to_gpu(cx)
    cy_saved_gpu = gpuarray.to_gpu(cy.transpose().copy())

    phi_gpu = gpuarray.to_gpu(phi)

    # Use this memory buffer for temporary c arrays and transposition
    buff_gpu = gpuarray.empty_like(cx_saved_gpu)

    # Calculate necessary buffer size
    bsize_int = cusparseDgtsvInterleavedBatch_bufferSizeExt(
        cusparse_handle, 0, L, ax_gpu.gpudata, bx_gpu.gpudata,
        buff_gpu.gpudata, phi_gpu.gpudata, L)
    pBuffer = pycuda.driver.mem_alloc(bsize_int)

    last_dt = np.inf
    while current_t < T:
        this_dt = min(dt, T - current_t)

        val10, val01 = _inject_mutations_2D_valcalc(this_dt, xx, xx, theta0, frozen1, frozen2,
                                                    nomut1, nomut2)
        # Use a single thread to update the two necessary values in phi
        kernels._inject_mutations_2D_vals(phi_gpu, L,
                                  val01, val10, block=(1,1,1))

        if not frozen1:
            # Restore from saved version of dux
            pycuda.driver.memcpy_dtod(buff_gpu.gpudata, cx_saved_gpu.gpudata,
                                      cx_saved_gpu.nbytes)

            # Prepare linear system
            bx_gpu += (1./this_dt - 1./last_dt)
            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                ax_gpu.gpudata, bx_gpu.gpudata, buff_gpu.gpudata, phi_gpu.gpudata,
                L, pBuffer)

        if not frozen2:
            # We tranpose here and use InterleavedBatch (rather than StridedBatch)
            # because it is notably faster.
            transpose_gpuarray(phi_gpu, buff_gpu)
            # Switch which memory location is treated as buffer and phiT
            phiT_gpu, buff_gpu = buff_gpu, phi_gpu
            phiT_gpu /= this_dt

            pycuda.driver.memcpy_dtod(buff_gpu.gpudata, cy_saved_gpu.gpudata,
                                      cy_saved_gpu.nbytes)
            by_gpu += (1./this_dt - 1./last_dt)

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                ay_gpu.gpudata, by_gpu.gpudata, buff_gpu.gpudata, phiT_gpu.gpudata,
                L, pBuffer)
            transpose_gpuarray(phiT_gpu, buff_gpu)
            phi_gpu, buff_gpu = buff_gpu, phiT_gpu

        last_dt = this_dt
        current_t += this_dt

    phi = phi_gpu.get().reshape(L,L)
    return phi

def _two_pops_temporal_params(phi, xx, T, initial_t, nu1_f, nu2_f, m12_f, m21_f, gamma1_f,
            gamma2_f, h1_f, h2_f, theta0_f, frozen1, frozen2, nomut1, nomut2):
    current_t = initial_t
    nu1,nu2 = nu1_f(current_t), nu2_f(current_t)
    m12,m21 = m12_f(current_t), m21_f(current_t)
    gamma1,gamma2 = gamma1_f(current_t), gamma2_f(current_t)
    h1,h2 = h1_f(current_t), h2_f(current_t)

    if dadi.Integration.use_delj_trick:
        raise ValueError("delj trick not currently supported in CUDA execution")

    yy = xx
    dx = dy = np.diff(xx)
    dfactor = dadi.Integration._compute_dfactor(dx)
    xInt = (xx[:-1] + xx[1:])*0.5

    L = M = np.int32(len(xx))

    xx_gpu = yy_gpu = gpuarray.to_gpu(xx)
    dx_gpu = gpuarray.to_gpu(dx)
    dfactor_gpu = gpuarray.to_gpu(dfactor)
    xInt_gpu = gpuarray.to_gpu(xInt)

    # By transposing phi, we can use the same functions to generate
    # the a,b,c matrices for both x and y.
    phi_gpu = gpuarray.to_gpu(phi)

    Vx_gpu = gpuarray.empty(L, np.float64)
    VIntx_gpu = gpuarray.empty(L-1, np.float64)

    ax_gpu = gpuarray.empty((L,M), np.float64)
    bx_gpu = gpuarray.empty((L,M), np.float64)
    # Buffer for MInt arrays, c arrays, and transposition
    buff_gpu = gpuarray.empty((L,M), np.float64)

    bsize_int = cusparseDgtsvInterleavedBatch_bufferSizeExt(
        cusparse_handle, 0, L, ax_gpu.gpudata, bx_gpu.gpudata,
        buff_gpu.gpudata, phi_gpu.gpudata, L)
    pBuffer = pycuda.driver.mem_alloc(bsize_int)

    while current_t < T:
        dt = min(dadi.Integration._compute_dt(dx,nu1,[m12],gamma1,h1),
                 dadi.Integration._compute_dt(dy,nu2,[m21],gamma2,h2))
        this_dt = np.float64(min(dt, T - current_t))

        next_t = current_t + this_dt

        nu1,nu2 = nu1_f(next_t), nu2_f(next_t)
        m12,m21 = m12_f(next_t), m21_f(next_t)
        gamma1,gamma2 = gamma1_f(next_t), gamma2_f(next_t)
        h1,h2 = h1_f(next_t), h2_f(next_t)
        theta0 = theta0_f(next_t)

        val10, val01 = dadi.cuda.Integration._inject_mutations_2D_valcalc(this_dt, xx, yy, theta0, frozen1, frozen2,
                                                                          nomut1, nomut2)
        kernels._inject_mutations_2D_vals(phi_gpu, L, np.float64(val01), np.float64(val10), 
                                          block=(1, 1, 1))

        if not frozen1:
            # Note, if L is an int32, L-1 can be int64. This can
            # break the code. So we wrap the -1 expressions in np.int32.
            kernels._Vfunc(xx_gpu, nu1, L, Vx_gpu, 
                           grid=_grid(L), block=_block())
            kernels._Vfunc(xInt_gpu, nu1, np.int32(L-1), VIntx_gpu, 
                           grid=_grid(L-1), block=_block())
            # Fill buff_gpu with MInt
            kernels._Mfunc2D(xInt_gpu, yy_gpu, m12, gamma1, h1,
                             np.int32(L-1), M, buff_gpu,
                             grid=_grid((L-1)*M), block=_block())

            bx_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                buff_gpu, Vx_gpu, this_dt, L, M,
                ax_gpu, bx_gpu,
                grid=_grid((L-1)*M), block=_block())
            # Note that this transforms buff_gpu from MInt to cx
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                buff_gpu, Vx_gpu, this_dt, L, M,
                bx_gpu, buff_gpu,
                grid=_grid((L-1)*M), block=_block())
            kernels._include_bc(dx_gpu, nu1, gamma1, h1, L, M,
                bx_gpu, block=(1,1,1))
            kernels._cx0(buff_gpu, L, M, grid=_grid(M), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                ax_gpu.gpudata, bx_gpu.gpudata, buff_gpu.gpudata, phi_gpu.gpudata,
                L, pBuffer)
        if not frozen2:
            # Use the buffer as destination for transpose
            transpose_gpuarray(phi_gpu, buff_gpu)
            # Use previous phi memory as buffer for this step, and vice versa
            phiT_gpu, buff_gpu = buff_gpu, phi_gpu

            phiT_gpu /= this_dt

            kernels._Vfunc(xx_gpu, nu2, L, Vx_gpu, 
                           grid=_grid(L), block=_block())
            kernels._Vfunc(xInt_gpu, nu2, np.int32(L-1), VIntx_gpu, 
                           grid=_grid(L-1), block=_block())
            kernels._Mfunc2D(xInt_gpu, yy_gpu, m21, gamma2, h2,
                             np.int32(L-1), M, buff_gpu,
                             grid=_grid((L-1)*M), block=_block())

            bx_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                buff_gpu, Vx_gpu, this_dt, L, M,
                ax_gpu, bx_gpu,
                grid=_grid((L-1)*M), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                buff_gpu, Vx_gpu, this_dt, L, M,
                bx_gpu, buff_gpu,
                grid=_grid((L-1)*M), block=_block())
            kernels._include_bc(dx_gpu, nu2, gamma2, h2, L, M,
                bx_gpu, block=(1,1,1))
            kernels._cx0(buff_gpu, L, M, grid=_grid(M), block=_block())

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                ax_gpu.gpudata, bx_gpu.gpudata, buff_gpu.gpudata, phiT_gpu.gpudata,
                L, pBuffer)
            transpose_gpuarray(phiT_gpu, buff_gpu)
            phi_gpu, buff_gpu = buff_gpu, phiT_gpu

        current_t = next_t

    return phi_gpu.get()

def _three_pops_const_params(phi, xx,
                theta0, frozen1, frozen2, frozen3, 
                ax, bx, cx, ay, by, cy, az, bz, cz,
                current_t, dt, T):

    L = np.int32(len(xx))

    ax_gpu = gpuarray.to_gpu(ax.reshape(L,L**2))
    bx_gpu = gpuarray.to_gpu(bx.reshape(L,L**2))
    cx_saved_gpu = gpuarray.to_gpu(cx.reshape(L,L**2))
    c_gpu = gpuarray.to_gpu(cx_saved_gpu)

    ay_gpu = gpuarray.to_gpu(ay.reshape(L,L**2).transpose().copy())
    by_gpu = gpuarray.to_gpu(by.reshape(L,L**2).transpose().copy())
    cy_saved_gpu = gpuarray.to_gpu(cy.reshape(L,L**2).transpose().copy())

    # There is probably a cleaner way to write this, but this matches
    # the transformations we'll be making to phi.
    az_gpu = gpuarray.to_gpu(az.reshape(L,L**2).transpose().reshape(L,L**2).transpose().copy())
    bz_gpu = gpuarray.to_gpu(bz.reshape(L,L**2).transpose().reshape(L,L**2).transpose().copy())
    cz_saved_gpu = gpuarray.to_gpu(cz.reshape(L,L**2).transpose().reshape(L,L**2).transpose().copy())

    # To save memory, in this 3D method we'll be using the c_gpu matrix
    # as the buffer for transposing.
    phi_gpu = gpuarray.to_gpu(phi.reshape(L,L**2))

    bsize_int = cusparseDgtsvInterleavedBatch_bufferSizeExt(
        cusparse_handle, 0, L, ax_gpu.gpudata, bx_gpu.gpudata,
        cx_saved_gpu.gpudata, phi_gpu.gpudata, L*L)
    pBuffer = pycuda.driver.mem_alloc(bsize_int)

    last_dt = np.inf
    while current_t < T:    
        this_dt = min(dt, T - current_t)

        val100, val010, val001 = \
            _inject_mutations_3D_valcalc(this_dt, xx, xx, xx, theta0, 
                                         frozen1, frozen2, frozen3)
        kernels._inject_mutations_3D_vals(phi_gpu, L,
                                          val001, val010, val100, block=(1,1,1))
        if not frozen1:
            pycuda.driver.memcpy_dtod(c_gpu.gpudata, cx_saved_gpu.gpudata,
                                      cx_saved_gpu.nbytes)

            bx_gpu += (1./this_dt - 1./last_dt)
            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                ax_gpu.gpudata, bx_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L**2, pBuffer)

        # We transpose into the c_gpu memory buffer.
        # Then we switch the names of the phi_gpu and c_gpu buffers.
        transpose_gpuarray(phi_gpu, c_gpu.reshape(L**2,L))
        phi_gpu, c_gpu = c_gpu.reshape(L,L**2), phi_gpu
        if not frozen2:
            pycuda.driver.memcpy_dtod(c_gpu.gpudata, cy_saved_gpu.gpudata,
                                      cy_saved_gpu.nbytes)
            by_gpu += (1./this_dt - 1./last_dt)
            phi_gpu /= this_dt
            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                ay_gpu.gpudata, by_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L**2, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L**2,L))
        phi_gpu, c_gpu = c_gpu.reshape(L,L**2), phi_gpu
        if not frozen3:
            pycuda.driver.memcpy_dtod(c_gpu.gpudata, cz_saved_gpu.gpudata,
                                      cz_saved_gpu.nbytes)
            bz_gpu += (1./this_dt - 1./last_dt)
            phi_gpu /= this_dt
            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                az_gpu.gpudata, bz_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L**2, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L**2,L))
        phi_gpu, c_gpu = c_gpu.reshape(L,L**2), phi_gpu

        last_dt = this_dt
        current_t += this_dt

    return phi_gpu.get().reshape(L,L,L)

def _three_pops_temporal_params(phi, xx, T, initial_t, nu1_f, nu2_f, nu3_f, 
            m12_f, m13_f, m21_f, m23_f, m31_f, m32_f, 
            gamma1_f, gamma2_f, gamma3_f, h1_f, h2_f, h3_f,
            theta0_f, frozen1, frozen2, frozen3):
    if dadi.Integration.use_delj_trick:
        raise ValueError("delj trick not currently supported in CUDA execution")

    t = current_t = initial_t
    nu1,nu2,nu3 = nu1_f(t), nu2_f(t), nu3_f(t)
    m12,m13,m21,m23,m31,m32 = m12_f(t), m13_f(t), m21_f(t), m23_f(t), m31_f(t), m32_f(t)
    gamma1,gamma2,gamma3 = gamma1_f(t), gamma2_f(t), gamma3_f(t)
    h1,h2,h3 = h1_f(t), h2_f(t), h3_f(t)

    L = M = N = np.int32(len(xx))

    phi_gpu = gpuarray.to_gpu(phi.reshape(L,M*N))

    yy = zz = xx
    dx = dy = dz = np.diff(xx)
    dfactor = dadi.Integration._compute_dfactor(dx)
    xInt = (xx[:-1] + xx[1:])*0.5

    xx_gpu = gpuarray.to_gpu(xx)
    dx_gpu = gpuarray.to_gpu(dx)
    dfactor_gpu = gpuarray.to_gpu(dfactor)
    xInt_gpu = gpuarray.to_gpu(xInt)

    V_gpu = gpuarray.empty(L, np.float64)
    VInt_gpu = gpuarray.empty(L-1, np.float64)

    a_gpu = gpuarray.empty((L,L*L), np.float64)
    b_gpu = gpuarray.empty((L,L*L), np.float64)
    c_gpu = gpuarray.empty((L,L*L), np.float64)

    bsize_int = cusparseDgtsvInterleavedBatch_bufferSizeExt(
        cusparse_handle, 0, L, a_gpu.gpudata, b_gpu.gpudata,
        c_gpu.gpudata, phi_gpu.gpudata, L**2)
    pBuffer = pycuda.driver.mem_alloc(bsize_int)

    while current_t < T:
        dt = min(dadi.Integration._compute_dt(dx, nu1, [m12, m13], gamma1, h1),
                 dadi.Integration._compute_dt(dy, nu2, [m21, m23], gamma2, h2),
                 dadi.Integration._compute_dt(dz, nu3, [m31, m32], gamma3, h3))
        this_dt = np.float64(min(dt, T - current_t))

        next_t = current_t + this_dt

        nu1,nu2,nu3 = nu1_f(next_t), nu2_f(next_t), nu3_f(next_t)
        m12,m13 = m12_f(next_t), m13_f(next_t)
        m21,m23 = m21_f(next_t), m23_f(next_t)
        m31,m32 = m31_f(next_t), m32_f(next_t)
        gamma1,gamma2 = gamma1_f(next_t), gamma2_f(next_t)
        gamma3 = gamma3_f(next_t)
        h1,h2,h3 = h1_f(next_t), h2_f(next_t), h3_f(next_t)
        theta0 = theta0_f(next_t)

        if np.any(np.less([T,nu1,nu2,nu3,m12,m13,m21,m23,m31,m32,theta0], 0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if np.any(np.equal([nu1,nu2,nu3], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        val100, val010, val001 = \
            _inject_mutations_3D_valcalc(this_dt, xx, yy, zz, theta0, 
                                         frozen1, frozen2, frozen3)
        kernels._inject_mutations_3D_vals(phi_gpu, L,
                                          val001, val010, val100, block=(1,1,1))
        # I can use the c_gpu buffer for the MInt_gpu buffer, to save GPU memory.
        # Note that I have to reassign this after each transpose operation I do.
        MInt_gpu = c_gpu
        if not frozen1:
            kernels._Vfunc(xx_gpu, nu1, L, V_gpu, 
                           grid=_grid(L), block=_block())
            kernels._Vfunc(xInt_gpu, nu1, np.int32(L-1), VInt_gpu, 
                           grid=_grid(L-1), block=_block())
            kernels._Mfunc3D(xInt_gpu, xx_gpu, xx_gpu, m12, m13, gamma1, h1,
                             np.int32(L-1), M, N, MInt_gpu,
                             grid=_grid((L-1)*M*N), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, L, M*N,
                a_gpu, b_gpu,
                grid=_grid((L-1)*M*N), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, L, M*N,
                b_gpu, c_gpu,
                grid=_grid((L-1)*M*N), block=_block())
            kernels._include_bc(dx_gpu, nu1, gamma1, h1, L, M*N,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, L, M*N, grid=_grid(M*N), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                M*N, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(M*N,L))
        phi_gpu, c_gpu = c_gpu.reshape(M,L*N), phi_gpu.reshape(M,L*N)
        MInt_gpu = c_gpu
        if not frozen2:
            kernels._Vfunc(xx_gpu, nu2, M, V_gpu, 
                           grid=_grid(M), block=_block())
            kernels._Vfunc(xInt_gpu, nu2, np.int32(M-1), VInt_gpu, 
                           grid=_grid(M-1), block=_block())
            # Note the order of the m23, m21 arguments here.
            kernels._Mfunc3D(xInt_gpu, xx_gpu, xx_gpu, m23, m21, gamma2, h2,
                             np.int32(M-1), N, L, MInt_gpu,
                             grid=_grid((M-1)*L*N), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, M, L*N,
                a_gpu, b_gpu,
                grid=_grid((M-1)*L*N), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, M, L*N,
                b_gpu, c_gpu,
                grid=_grid((M-1)*L*N), block=_block())
            kernels._include_bc(dx_gpu, nu2, gamma2, h2, M, L*N,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, M, L*N, grid=_grid(L*N), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, M,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*N, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*N,M))
        phi_gpu, c_gpu = c_gpu.reshape(N,L*M), phi_gpu.reshape(N,L*M)
        MInt_gpu = c_gpu
        if not frozen3:
            kernels._Vfunc(xx_gpu, nu3, N, V_gpu, 
                           grid=_grid(N), block=_block())
            kernels._Vfunc(xInt_gpu, nu3, np.int32(N-1), VInt_gpu, 
                           grid=_grid(N-1), block=_block())
            kernels._Mfunc3D(xInt_gpu, xx_gpu, xx_gpu, m31, m32, gamma3, h3,
                             np.int32(N-1), L, M, MInt_gpu,
                             grid=_grid((N-1)*M*L), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, N, L*M,
                a_gpu, b_gpu,
                grid=_grid((N-1)*L*M), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, N, L*M,
                b_gpu, c_gpu,
                grid=_grid((N-1)*L*M), block=_block())
            kernels._include_bc(dx_gpu, nu3, gamma3, h3, N, L*M,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, N, L*M, grid=_grid(L*M), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, N,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*M, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(M*N,L))
        phi_gpu, c_gpu = c_gpu.reshape(L,M*N), phi_gpu.reshape(L,M*N)

        current_t += this_dt

    return phi_gpu.get().reshape(L,M,N)

def _four_pops_temporal_params(phi, xx, T, initial_t, nu1_f, nu2_f, nu3_f, nu4_f,
            m12_f, m13_f, m14_f, m21_f, m23_f, m24_f, m31_f, m32_f, m34_f,
            m41_f, m42_f, m43_f, gamma1_f, gamma2_f, gamma3_f, gamma4_f,
            h1_f, h2_f, h3_f, h4_f, theta0_f, frozen1, frozen2, frozen3, frozen4):
    if dadi.Integration.use_delj_trick:
        raise ValueError("delj trick not currently supported in CUDA execution")

    current_t = initial_t
    nu1, nu2, nu3, nu4 = nu1_f(current_t), nu2_f(current_t), nu3_f(current_t), nu4_f(current_t)
    gamma1, gamma2, gamma3, gamma4 = gamma1_f(current_t), gamma2_f(current_t), gamma3_f(current_t), gamma4_f(current_t)
    h1, h2, h3, h4 = h1_f(current_t), h2_f(current_t), h3_f(current_t), h4_f(current_t)
    m12, m13, m14 = m12_f(current_t), m13_f(current_t), m14_f(current_t)
    m21, m23, m24 = m21_f(current_t), m23_f(current_t), m24_f(current_t)
    m31, m32, m34 = m31_f(current_t), m32_f(current_t), m34_f(current_t)
    m41, m42, m43 = m41_f(current_t), m42_f(current_t), m43_f(current_t)

    L = M = N = O = np.int32(len(xx))

    phi_gpu = gpuarray.to_gpu(phi.reshape(L,M*N*O))

    aa = yy = zz = xx
    da = dx = dy = dz = np.diff(xx)
    dfactor = dadi.Integration._compute_dfactor(dx)
    xInt = (xx[:-1] + xx[1:])*0.5

    xx_gpu = gpuarray.to_gpu(xx)
    dx_gpu = gpuarray.to_gpu(dx)
    dfactor_gpu = gpuarray.to_gpu(dfactor)
    xInt_gpu = gpuarray.to_gpu(xInt)

    V_gpu = gpuarray.empty(L, np.float64)
    VInt_gpu = gpuarray.empty(L-1, np.float64)

    a_gpu = gpuarray.empty((L,M*N*O), np.float64)
    b_gpu = gpuarray.empty((L,M*N*O), np.float64)
    c_gpu = gpuarray.empty((L,M*N*O), np.float64)

    bsize_int = cusparseDgtsvInterleavedBatch_bufferSizeExt(
        cusparse_handle, 0, L, a_gpu.gpudata, b_gpu.gpudata,
        c_gpu.gpudata, phi_gpu.gpudata, M*N*O)
    pBuffer = pycuda.driver.mem_alloc(bsize_int)

    while current_t < T:
        dt = min(dadi.Integration._compute_dt(dx, nu1, [m12, m13, m14], gamma1, h1),
                 dadi.Integration._compute_dt(dy, nu2, [m21, m23, m24], gamma2, h2),
                 dadi.Integration._compute_dt(dz, nu3, [m31, m32, m34], gamma3, h3),
                 dadi.Integration._compute_dt(da, nu4, [m41, m42, m43], gamma4, h4))
        this_dt = np.float64(min(dt, T - current_t))

        next_t = current_t + this_dt

        nu1, nu2, nu3, nu4 = nu1_f(next_t), nu2_f(next_t), nu3_f(next_t), nu4_f(next_t)
        gamma1, gamma2, gamma3, gamma4 = gamma1_f(next_t), gamma2_f(next_t), gamma3_f(next_t), gamma4_f(next_t)
        h1, h2, h3, h4 = h1_f(next_t), h2_f(next_t), h3_f(next_t), h4_f(next_t)
        m12, m13, m14 = m12_f(next_t), m13_f(next_t), m14_f(next_t)
        m21, m23, m24 = m21_f(next_t), m23_f(next_t), m24_f(next_t)
        m31, m32, m34 = m31_f(next_t), m32_f(next_t), m34_f(next_t)
        m41, m42, m43 = m41_f(next_t), m42_f(next_t), m43_f(next_t)
        theta0 = theta0_f(next_t)

        if np.any(np.less([T,nu1,nu2,nu3,nu4,m12,m13,m14,m21,m23,m24,m31,m32,m34,m41,m42,m43,theta0], 0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if np.any(np.equal([nu1,nu2,nu3,nu4], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        val1000, val0100, val0010, val0001 = \
            _inject_mutations_4D_valcalc(this_dt, xx, yy, zz, aa, theta0, 
                                         frozen1, frozen2, frozen3, frozen4)
        kernels._inject_mutations_4D_vals(phi_gpu, L,
                                          val0001, val0010, val0100, val1000, block=(1,1,1))
        # I can use the c_gpu buffer for the MInt_gpu buffer, to save GPU memory.
        # Note that I have to reassign this after each transpose operation I do.
        MInt_gpu = c_gpu
        if not frozen1:
            kernels._Vfunc(xx_gpu, nu1, L, V_gpu, 
                           grid=_grid(L), block=_block())
            kernels._Vfunc(xInt_gpu, nu1, np.int32(L-1), VInt_gpu, 
                           grid=_grid(L-1), block=_block())
            kernels._Mfunc4D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, m12, m13, m14, gamma1, h1,
                             np.int32(L-1), M, N, O, MInt_gpu,
                             grid=_grid((L-1)*M*N*O), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, L, M*N*O,
                a_gpu, b_gpu,
                grid=_grid((L-1)*M*N*O), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, L, M*N*O,
                b_gpu, c_gpu,
                grid=_grid((L-1)*M*N*O), block=_block())
            kernels._include_bc(dx_gpu, nu1, gamma1, h1, L, M*N*O,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, L, M*N*O, grid=_grid(M*N*O), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                M*N*O, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(M*N*O,L))
        phi_gpu, c_gpu = c_gpu.reshape(M,L*N*O), phi_gpu.reshape(M,L*N*O)
        MInt_gpu = c_gpu
        if not frozen2:
            kernels._Vfunc(xx_gpu, nu2, M, V_gpu, 
                           grid=_grid(M), block=_block())
            kernels._Vfunc(xInt_gpu, nu2, np.int32(M-1), VInt_gpu, 
                           grid=_grid(M-1), block=_block())
            # Note the order of the m arguments here.
            kernels._Mfunc4D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, m23, m24, m21, gamma2, h2,
                             np.int32(M-1), N, O, L, MInt_gpu,
                             grid=_grid((M-1)*L*N*O), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, M, L*N*O,
                a_gpu, b_gpu,
                grid=_grid((M-1)*L*N*O), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, M, L*N*O,
                b_gpu, c_gpu,
                grid=_grid((M-1)*L*N*O), block=_block())
            kernels._include_bc(dx_gpu, nu2, gamma2, h2, M, L*N*O,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, M, L*N*O, grid=_grid(L*N*O), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, M,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*N*O, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*N*O,M))
        phi_gpu, c_gpu = c_gpu.reshape(N,L*M*O), phi_gpu.reshape(N,L*M*O)
        MInt_gpu = c_gpu
        if not frozen3:
            kernels._Vfunc(xx_gpu, nu3, N, V_gpu, 
                           grid=_grid(N), block=_block())
            kernels._Vfunc(xInt_gpu, nu3, np.int32(N-1), VInt_gpu, 
                           grid=_grid(N-1), block=_block())
            kernels._Mfunc4D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, m34, m31, m32, gamma3, h3,
                             np.int32(N-1), O, L, M, MInt_gpu,
                             grid=_grid((N-1)*M*L*O), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, N, L*M*O,
                a_gpu, b_gpu,
                grid=_grid((N-1)*L*M*O), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, N, L*M*O,
                b_gpu, c_gpu,
                grid=_grid((N-1)*L*M*O), block=_block())
            kernels._include_bc(dx_gpu, nu3, gamma3, h3, N, L*M*O,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, N, L*M*O, grid=_grid(L*M*O), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, N,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*M*O, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*M*O,N))
        phi_gpu, c_gpu = c_gpu.reshape(O,L*M*N), phi_gpu.reshape(O,L*M*N)
        MInt_gpu = c_gpu
        if not frozen4:
            kernels._Vfunc(xx_gpu, nu4, O, V_gpu, 
                           grid=_grid(O), block=_block())
            kernels._Vfunc(xInt_gpu, nu4, np.int32(O-1), VInt_gpu, 
                           grid=_grid(O-1), block=_block())
            kernels._Mfunc4D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, m41, m42, m43, gamma4, h4,
                             np.int32(O-1), L, M, N, MInt_gpu,
                             grid=_grid((O-1)*M*L*N), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, O, L*M*N,
                a_gpu, b_gpu,
                grid=_grid((O-1)*L*M*N), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, O, L*M*N,
                b_gpu, c_gpu,
                grid=_grid((O-1)*L*M*N), block=_block())
            kernels._include_bc(dx_gpu, nu4, gamma4, h4, O, L*M*N,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, O, L*M*N, grid=_grid(L*M*N), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, O,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*M*N, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*M*N,O))
        phi_gpu, c_gpu = c_gpu.reshape(L,M*N*O), phi_gpu.reshape(L,M*N*O)

        current_t += this_dt

    return phi_gpu.get().reshape(L,M,N,O)

def _five_pops_temporal_params(phi, xx, T, initial_t, nu1_f, nu2_f, nu3_f, nu4_f, nu5_f,
            m12_f, m13_f, m14_f, m15_f, m21_f, m23_f, m24_f, m25_f, m31_f, m32_f, m34_f, m35_f,
            m41_f, m42_f, m43_f, m45_f, m51_f, m52_f, m53_f, m54_f, 
            gamma1_f, gamma2_f, gamma3_f, gamma4_f, gamma5_f,
            h1_f, h2_f, h3_f, h4_f, h5_f, theta0_f, frozen1, frozen2, frozen3, frozen4, frozen5):
    if dadi.Integration.use_delj_trick:
        raise ValueError("delj trick not currently supported in CUDA execution")

    current_t = initial_t
    nu1, nu2, nu3, nu4, nu5 = nu1_f(current_t), nu2_f(current_t), nu3_f(current_t), nu4_f(current_t), nu5_f(current_t)
    gamma1, gamma2, gamma3, gamma4, gamma5 = gamma1_f(current_t), gamma2_f(current_t), gamma3_f(current_t), gamma4_f(current_t), gamma5_f(current_t)
    h1, h2, h3, h4, h5 = h1_f(current_t), h2_f(current_t), h3_f(current_t), h4_f(current_t), h5_f(current_t)
    m12, m13, m14, m15 = m12_f(current_t), m13_f(current_t), m14_f(current_t), m15_f(current_t)
    m21, m23, m24, m25 = m21_f(current_t), m23_f(current_t), m24_f(current_t), m25_f(current_t)
    m31, m32, m34, m35 = m31_f(current_t), m32_f(current_t), m34_f(current_t), m35_f(current_t)
    m41, m42, m43, m45 = m41_f(current_t), m42_f(current_t), m43_f(current_t), m45_f(current_t)
    m51, m52, m53, m54 = m51_f(current_t), m52_f(current_t), m53_f(current_t), m54_f(current_t)
    nu1, nu2, nu3, nu4 = nu1_f(current_t), nu2_f(current_t), nu3_f(current_t), nu4_f(current_t)

    L = M = N = O = P = np.int32(len(xx))

    phi_gpu = gpuarray.to_gpu(phi.reshape(L,M*N*O*P))

    bb = aa = yy = zz = xx
    db = da = dx = dy = dz = np.diff(xx)
    dfactor = dadi.Integration._compute_dfactor(dx)
    xInt = (xx[:-1] + xx[1:])*0.5

    xx_gpu = gpuarray.to_gpu(xx)
    dx_gpu = gpuarray.to_gpu(dx)
    dfactor_gpu = gpuarray.to_gpu(dfactor)
    xInt_gpu = gpuarray.to_gpu(xInt)

    V_gpu = gpuarray.empty(L, np.float64)
    VInt_gpu = gpuarray.empty(L-1, np.float64)

    a_gpu = gpuarray.empty((L,M*N*O*P), np.float64)
    b_gpu = gpuarray.empty((L,M*N*O*P), np.float64)
    c_gpu = gpuarray.empty((L,M*N*O*P), np.float64)

    bsize_int = cusparseDgtsvInterleavedBatch_bufferSizeExt(
        cusparse_handle, 0, L, a_gpu.gpudata, b_gpu.gpudata,
        c_gpu.gpudata, phi_gpu.gpudata, M*N*O*P)
    pBuffer = pycuda.driver.mem_alloc(bsize_int)

    while current_t < T:
        dt = min(dadi.Integration._compute_dt(dx, nu1, [m12,m13,m14,m15], gamma1, h1),
                 dadi.Integration._compute_dt(dy, nu2, [m21,m23,m24,m25], gamma2, h2),
                 dadi.Integration._compute_dt(dz, nu3, [m31,m32,m34,m35], gamma3, h3),
                 dadi.Integration._compute_dt(da, nu4, [m41,m42,m43,m45], gamma4, h4),
                 dadi.Integration._compute_dt(db, nu5, [m51,m52,m53,m54],gamma5,h5))
        this_dt = np.float64(min(dt, T - current_t))

        next_t = current_t + this_dt

        nu1, nu2, nu3, nu4, nu5 = nu1_f(next_t), nu2_f(next_t), nu3_f(next_t), nu4_f(next_t), nu5_f(next_t)
        gamma1, gamma2, gamma3, gamma4, gamma5 = gamma1_f(next_t), gamma2_f(next_t), gamma3_f(next_t), gamma4_f(next_t), gamma5_f(next_t)
        h1, h2, h3, h4, h5 = h1_f(next_t), h2_f(next_t), h3_f(next_t), h4_f(next_t), h5_f(next_t)
        m12, m13, m14, m15 = m12_f(next_t), m13_f(next_t), m14_f(next_t), m15_f(next_t)
        m21, m23, m24, m25 = m21_f(next_t), m23_f(next_t), m24_f(next_t), m25_f(next_t)
        m31, m32, m34, m35 = m31_f(next_t), m32_f(next_t), m34_f(next_t), m35_f(next_t)
        m41, m42, m43, m45 = m41_f(next_t), m42_f(next_t), m43_f(next_t), m45_f(next_t)
        m51, m52, m53, m54 = m51_f(next_t), m52_f(next_t), m53_f(next_t), m54_f(next_t)
        theta0 = theta0_f(next_t)

        if np.any(np.less([T,nu1,nu2,nu3,nu4,nu5,m12,m13,m14,m15,m21,
                                 m23,m24,m25, m31,m32,m34,m35, m41,m42,m43,m45,
                                 m51,m52,m53,m54, theta0],
                                0)):
            raise ValueError('A time, population size, migration rate, or '
                             'theta0 is < 0. Has the model been mis-specified?')
        if np.any(np.equal([nu1,nu2,nu3,nu4,nu5], 0)):
            raise ValueError('A population size is 0. Has the model been '
                             'mis-specified?')

        val10000, val01000, val00100, val00010, val00001 = \
            _inject_mutations_5D_valcalc(this_dt, xx, yy, zz, aa, bb, theta0, 
                                         frozen1, frozen2, frozen3, frozen4, frozen5)
        kernels._inject_mutations_5D_vals(phi_gpu, L,
                                          val00001, val00010, val00100, val01000, val10000, block=(1,1,1))
        # I can use the c_gpu buffer for the MInt_gpu buffer, to save GPU memory.
        # Note that I have to reassign this after each transpose operation I do.
        MInt_gpu = c_gpu
        if not frozen1:
            kernels._Vfunc(xx_gpu, nu1, L, V_gpu, 
                           grid=_grid(L), block=_block())
            kernels._Vfunc(xInt_gpu, nu1, np.int32(L-1), VInt_gpu, 
                           grid=_grid(L-1), block=_block())
            kernels._Mfunc5D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, xx_gpu, m12, m13, m14, m15, gamma1, h1,
                             np.int32(L-1), M, N, O, P, MInt_gpu,
                             grid=_grid((L-1)*M*N*O*P), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, L, M*N*O*P,
                a_gpu, b_gpu,
                grid=_grid((L-1)*M*N*O*P), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, L, M*N*O*P,
                b_gpu, c_gpu,
                grid=_grid((L-1)*M*N*O*P), block=_block())
            kernels._include_bc(dx_gpu, nu1, gamma1, h1, L, M*N*O*P,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, L, M*N*O*P, grid=_grid(M*N*O*P), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, L,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                M*N*O*P, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(M*N*O*P,L))
        phi_gpu, c_gpu = c_gpu.reshape(M,L*N*O*P), phi_gpu.reshape(M,L*N*O*P)
        MInt_gpu = c_gpu
        if not frozen2:
            kernels._Vfunc(xx_gpu, nu2, M, V_gpu, 
                           grid=_grid(M), block=_block())
            kernels._Vfunc(xInt_gpu, nu2, np.int32(M-1), VInt_gpu, 
                           grid=_grid(M-1), block=_block())
            # Note the order of the m arguments here.
            kernels._Mfunc5D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, xx_gpu, m23, m24, m25, m21, gamma2, h2,
                             np.int32(M-1), N, O, P, L, MInt_gpu,
                             grid=_grid((M-1)*L*N*O*P), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, M, L*N*O*P,
                a_gpu, b_gpu,
                grid=_grid((M-1)*L*N*O*P), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, M, L*N*O*P,
                b_gpu, c_gpu,
                grid=_grid((M-1)*L*N*O*P), block=_block())
            kernels._include_bc(dx_gpu, nu2, gamma2, h2, M, L*N*O*P,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, M, L*N*O*P, grid=_grid(L*N*O*P), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, M,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*N*O*P, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*N*O*P,M))
        phi_gpu, c_gpu = c_gpu.reshape(N,L*M*O*P), phi_gpu.reshape(N,L*M*O*P)
        MInt_gpu = c_gpu
        if not frozen3:
            kernels._Vfunc(xx_gpu, nu3, N, V_gpu, 
                           grid=_grid(N), block=_block())
            kernels._Vfunc(xInt_gpu, nu3, np.int32(N-1), VInt_gpu, 
                           grid=_grid(N-1), block=_block())
            kernels._Mfunc5D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, xx_gpu, m34, m35, m31, m32, gamma3, h3,
                             np.int32(N-1), O, P, L, M, MInt_gpu,
                             grid=_grid((N-1)*M*L*O*P), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, N, L*M*O*P,
                a_gpu, b_gpu,
                grid=_grid((N-1)*L*M*O*P), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, N, L*M*O*P,
                b_gpu, c_gpu,
                grid=_grid((N-1)*L*M*O*P), block=_block())
            kernels._include_bc(dx_gpu, nu3, gamma3, h3, N, L*M*O*P,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, N, L*M*O*P, grid=_grid(L*M*O*P), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, N,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*M*O*P, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*M*O*P,N))
        phi_gpu, c_gpu = c_gpu.reshape(O,L*M*N*P), phi_gpu.reshape(O,L*M*N*P)
        MInt_gpu = c_gpu
        if not frozen4:
            kernels._Vfunc(xx_gpu, nu4, O, V_gpu, 
                           grid=_grid(O), block=_block())
            kernels._Vfunc(xInt_gpu, nu4, np.int32(O-1), VInt_gpu, 
                           grid=_grid(O-1), block=_block())
            kernels._Mfunc5D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, xx_gpu, m45, m41, m42, m43, gamma4, h4,
                             np.int32(O-1), P, L, M, N, MInt_gpu,
                             grid=_grid((O-1)*M*L*N*P), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, O, L*M*N*P,
                a_gpu, b_gpu,
                grid=_grid((O-1)*L*M*N*P), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, O, L*M*N*P,
                b_gpu, c_gpu,
                grid=_grid((O-1)*L*M*N*P), block=_block())
            kernels._include_bc(dx_gpu, nu4, gamma4, h4, O, L*M*N*P,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, O, L*M*N*P, grid=_grid(L*M*N*P), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, O,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*M*N*P, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*M*N*P,O))
        phi_gpu, c_gpu = c_gpu.reshape(P,L*M*N*O), phi_gpu.reshape(P,L*M*N*O)
        MInt_gpu = c_gpu
        if not frozen5:
            kernels._Vfunc(xx_gpu, nu5, P, V_gpu, 
                           grid=_grid(P), block=_block())
            kernels._Vfunc(xInt_gpu, nu5, np.int32(P-1), VInt_gpu, 
                           grid=_grid(P-1), block=_block())
            kernels._Mfunc5D(xInt_gpu, xx_gpu, xx_gpu, xx_gpu, xx_gpu, m51, m52, m53, m54, gamma5, h5,
                             np.int32(P-1), L, M, N, O, MInt_gpu,
                             grid=_grid((P-1)*L*M*N*O), block=_block())

            b_gpu.fill(1./this_dt)
            kernels._compute_ab_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, P, L*M*N*O,
                a_gpu, b_gpu,
                grid=_grid((P-1)*L*M*N*O), block=_block())
            kernels._compute_bc_nobc(dx_gpu, dfactor_gpu, 
                MInt_gpu, V_gpu, this_dt, P, L*M*N*O,
                b_gpu, c_gpu,
                grid=_grid((P-1)*L*M*N*O), block=_block())
            kernels._include_bc(dx_gpu, nu5, gamma5, h5, P, L*M*N*O,
                b_gpu, block=(1,1,1))
            kernels._cx0(c_gpu, P, L*M*N*O, grid=_grid(L*M*N*O), block=_block())

            phi_gpu /= this_dt

            cusparseDgtsvInterleavedBatch(cusparse_handle, 0, P,
                a_gpu.gpudata, b_gpu.gpudata, c_gpu.gpudata, phi_gpu.gpudata,
                L*M*N*O, pBuffer)

        transpose_gpuarray(phi_gpu, c_gpu.reshape(L*M*N*O,P))
        phi_gpu, c_gpu = c_gpu.reshape(L,M*N*O*P), phi_gpu.reshape(L,M*N*O*P)

        current_t += this_dt

    return phi_gpu.get().reshape(L,M,N,O,P)
